# SPDX-License-Identifier: BSD-2-Clause
import argparse
import os
import sys
import traceback
import yaml

# local imports
from .command import Command
from .command import commandlet

from .db import Db
from .utils import bytes_to_file
from .utils import TemporaryDirectory
from .utils import query_yes_no
from .utils import create_primary

from .tpm2 import Tpm2

@commandlet("init")
class InitCommand(Command):
    '''
    Initializes a tpm2-pkcs11 store
    '''

    # adhere to an interface
    # pylint: disable=no-self-use
    def generate_options(self, group_parser):
        mg = group_parser.add_mutually_exclusive_group()
        # Keep the old --owner-auth kicking around for backwards compat
        # but hide it in the help output.
        mg.add_argument(
            '--hierarchy-auth',
            help='The authorization password for adding a primary object to the hierarchy.\n',
            default="")
        mg.add_argument(
            '--owner-auth',
            help=argparse.SUPPRESS,
            dest='hierarchy_auth',
            default="")
        group_parser.add_argument(
            '--primary-auth',
            help='Authorization value for existing primary key object, defaults to an empty auth value.',
            default="")
        exclusive_group = group_parser.add_mutually_exclusive_group()
        exclusive_group.add_argument(
            '--primary-handle',
            nargs='?',
            type=InitCommand.str_to_handle,
            action=InitCommand.make_action(primary=True),
            help='Use an existing primary key object, defaults to 0x81000001.')
        exclusive_group.add_argument(
            '--transient-parent',
            const='tpm2-tools-default',
            nargs='?',
            choices=[ t for t in Tpm2.TEMPLATES.keys() if t is not None ],
            help='use a transient primary object of a given template.')

    @staticmethod
    def str_to_handle(arg):
        try:
            return int(arg, 0)
        except ValueError:
                if not os.path.exists(arg):
                    sys.exit('arg "%s" neither a handle nor esys handle file')
                else:
                    return arg


    @staticmethod
    def make_action(**kwargs):
        class customAction(argparse.Action):
            def __call__(self, parser, args, values, option_string=None):
                args.__dict__.update(kwargs)
                setattr(args, self.dest, values)

        return customAction

    def __call__(self, args):

        use_existing_primary = 'primary' in args and args['primary']

        path = args['path']
        if not os.path.isdir(path):
            sys.exit("Specified path is not a directory, got: %s" % (path))

        hierarchyauth = args['hierarchy_auth']
        pobjauth = args['primary_auth']

        # create the db
        with Db(path) as db:

            transient_parent = args['transient_parent']
            shall_evict = False
            with TemporaryDirectory() as d:
                try:
                    tpm2 = Tpm2(d)

                    if not use_existing_primary:
                        pobj_ctx = create_primary(tpm2, hierarchyauth, pobjauth, transient_parent)
                        if not transient_parent:
                            # Check if the default handle 0x81000001 is available
                            handle = 0x81000001
                            handles_persistent = yaml.safe_load(tpm2.getcap('handles-persistent'))
                            if handles_persistent and handle in handles_persistent:
                                # Automatically find the first free handle
                                handle = None

                            pobj_ctx = tpm2.evictcontrol(hierarchyauth, pobj_ctx, handle)
                            shall_evict = True
                    else:
                        # get the primary object auth value and convert it to hex
                        if pobjauth is None:
                            pobjauth = ""

                        handle = args['primary_handle'] if args['primary_handle'] is not None else 0x81000001

                        # If we get a raw handle, capture the ESYS_TR serialized and
                        # and verify that its persistent.
                        #
                        # TODO: with python bindings call esys_tr_from_public() over readpublic.
                        # TODO: we need a sapi handle from esys tr to look into getcap to see if
                        # its persistent.
                        if isinstance(handle, int):
                            (_, pobj_ctx) = tpm2.readpublic(handle)

                            # verify handle is persistent
                            output = tpm2.getcap('handles-persistent')
                            y = yaml.safe_load(output)
                            if handle not in y:
                                sys.exit('Handle 0x%x is not persistent' %
                                         (handle))
                        else:
                            pobj_ctx = handle

                    y = {
                        'transient' : bool(transient_parent)
                    }

                    if transient_parent:
                        y['template-name'] = args['transient_parent']
                    else:
                        with open(pobj_ctx, 'rb') as f:
                            y['esys-tr'] = bytes.hex(f.read())
                    pid = db.addprimary(y, pobjauth)

                    d = {
                        'id' : pid,
                        'action' : "Added" if use_existing_primary else "Created"
                    }

                    print(yaml.safe_dump(d, default_flow_style=False))

                except Exception as e:
                    if shall_evict and pobj_ctx != None:
                        tpm2.evictcontrol(hierarchyauth, pobj_ctx)

                    traceback.print_exc(file=sys.stdout)
                    sys.exit(e)


@commandlet("destroy")
class DestroyCommand(Command):
    '''
    Destroys a tpm2-pkcs11 store
    '''

    # adhere to an interface
    # pylint: disable=no-self-use
    def generate_options(self, group_parser):
        group_parser.add_argument(
            '--pid', type=int, help="The primary object id to remove.\n")
        group_parser.add_argument(
            '--hierarchy-auth',
            default="",
            help="The primary object id to remove.\n")

    def __call__(self, args):
        path = args['path']
        pid = args['pid']
        hierarchyauth = args['hierarchy_auth']

        if not os.path.exists(path):
            os.mkdir(path)
        elif not os.path.isdir(path):
            sys.exit("Specified path is not a directory, got: %s" % (path))

        proceed = query_yes_no(
            'This will delete the primary object of id "%s" and all associated data from db under "%s"'
            % (pid, path))
        if not proceed:
            sys.exit(0)
        # create the db
        with Db(path) as db:
            pobj = db.getprimary(pid)
            if pobj is None:
                sys.exit('Primary Object id "%s"not found' % pid)

            with TemporaryDirectory() as d:
                tpm2 = Tpm2(d)

                tr_file = bytes_to_file(pobj['handle'], d)

                db.rmprimary(pid)
                tpm2.evictcontrol(hierarchyauth, tr_file)

@commandlet("dbup")
class DbUp(Command):
    '''
    Initializes a tpm2-pkcs11 store
    '''

    # adhere to an interface
    # pylint: disable=no-self-use
    def generate_options(self, group_parser):
        pass

    def __call__(self, args):

        path = args['path']

        # create the db
        with Db(path) as db:
            old_ver = db.version
            new_ver = db.VERSION
            y = {
                'old' : old_ver,
                'new' : new_ver,
            }

            print(yaml.safe_dump(y, default_flow_style=False))



# SPDX-License-Identifier: BSD-2-Clause
import fcntl
import io
import os
import sys
import sqlite3
import textwrap
import yaml

from .pkcs11t import (
    CKA_ALLOWED_MECHANISMS,
    CKA_CLASS,
    CKO_SECRET_KEY,
    CKO_PRIVATE_KEY,
    CKA_KEY_TYPE,
    CKM_AES_CBC_PAD,
    CKM_AES_CTR,
    CKK_AES,
    CKK_EC,
    CKM_ECDSA_SHA256,
    CKM_ECDSA_SHA384,
    CKM_ECDSA_SHA512
)

VERSION = 8

#
# With Db() as db:
# // do stuff
#
class Db(object):
    def __init__(self, dirpath):
        self._path = os.path.join(dirpath, "tpm2_pkcs11.sqlite3")

    def __enter__(self):
        self._conn = sqlite3.connect(self._path)
        self._conn.row_factory = sqlite3.Row
        self._conn.execute('PRAGMA foreign_keys = ON;')
        self._create()

        return self

    @staticmethod
    def _blobify(path):
        with open(path, 'rb') as f:
            ablob = f.read()
            return sqlite3.Binary(ablob)

    def gettoken(self, label=None, id=None):
        c = self._conn.cursor()
        if label != None:
            c.execute("SELECT * from tokens WHERE label=?", (label, ))
        elif id != None:
            c.execute("SELECT * from tokens WHERE id=?", (id, ))
        x = c.fetchone()
        if x is None:
            sys.exit('No token labeled "%s"' % label)
        return x

    def getsealobject(self, tokid):
        c = self._conn.cursor()
        c.execute("SELECT * from sealobjects WHERE tokid=?", (tokid, ))
        x = c.fetchone()
        return x

    def getprimaries(self):
        c = self._conn.cursor()
        c.execute("SELECT * from pobjects")
        x = c.fetchall()
        return x

    def gettokens(self, pid):
        c = self._conn.cursor()
        c.execute("SELECT * from tokens WHERE pid=?", (pid, ))
        x = c.fetchall()
        return x

    def getobjects(self, tokid):
        c = self._conn.cursor()
        c.execute("SELECT * from tobjects WHERE tokid=?", (tokid, ))
        x = c.fetchall()
        return x

    def rmtoken(self, label):
        # This works on the premise of a cascading delete tied by foreign
        # key relationships.
        self._conn.execute('DELETE from tokens WHERE label=?', (label, ))

    def getprimary(self, pid):
        c = self._conn.cursor()
        c.execute("SELECT * from pobjects WHERE id=?", (pid, ))
        x = c.fetchone()
        return x

    def rmprimary(self, pid):
        # This works on the premise of a cascading delete tied by foreign
        # key relationships.
        self._conn.execute('DELETE from pobjects WHERE id=?', (pid, ))

    def gettertiary(self, tokid):
        c = self._conn.cursor()
        c.execute("SELECT * from tobjects WHERE tokid=?", (tokid, ))
        x = c.fetchall()
        return x

    def getobject(self, tid):
        c = self._conn.cursor()
        c.execute("SELECT * from tobjects WHERE id=?", (tid, ))
        x = c.fetchone()
        return x

    def getpid_by_tokid(self, tokid):
        c = self._conn.cursor()
        c.execute("SELECT pid from tokens WHERE id=?", (tokid, ))
        x = c.fetchone()
        return x['pid']

    def rmobject(self, tid):
        c = self._conn.cursor()
        c.execute("DELETE FROM tobjects WHERE id=?", (tid, ))

    def addtoken(self, pid, config, label=None):

        token = {
            # General Metadata
            'pid': pid,
            'config': yaml.dump(config, canonical=True)
        }

        if 'token-init=True' in token['config'] and label is None:
            raise RuntimeError('Expected label if token is to be initialized')

        if label:
            token['label'] = label

        columns = ', '.join(token.keys())
        placeholders = ', '.join('?' * len(token))
        sql = 'INSERT INTO tokens ({}) VALUES ({})'.format(columns,
                                                           placeholders)
        c = self._conn.cursor()
        c.execute(sql, list(token.values()))

        return c.lastrowid

    def updateconfig(self, token, config):

        new_config = yaml.dump(config, canonical=True)

        sql = 'UPDATE tokens SET config=? WHERE id=?'

        values = (new_config, token['id'])

        c = self._conn.cursor()
        c.execute(sql, values)

    def addsealobjects(self, tokid, usersealauth, usersealpriv, usersealpub,
                       sosealauth, sosealpriv, sosealpub):

        sealobjects = {
            # General Metadata
            'tokid': tokid,
            'userpriv': Db._blobify(usersealpriv),
            'userpub': Db._blobify(usersealpub),
            'sopriv': Db._blobify(sosealpriv),
            'sopub': Db._blobify(sosealpub),
            'userauthsalt': usersealauth['salt'],
            'soauthsalt': sosealauth['salt'],
        }

        columns = ', '.join(sealobjects.keys())
        placeholders = ', '.join('?' * len(sealobjects))
        sql = 'INSERT INTO sealobjects ({}) VALUES ({})'.format(columns,
                                                                placeholders)
        c = self._conn.cursor()
        c.execute(sql, list(sealobjects.values()))

        return c.lastrowid

    @staticmethod
    def addprimary_raw(cursor, tablename, config, objauth, hierarchy='o', pid=None):

        # Subordinate commands will need some of this data
        # when deriving subordinate objects, so pass it back
        pobject = {
            # General Metadata
            'hierarchy': hierarchy,
            'config': yaml.safe_dump(config, canonical=True),
            'objauth': objauth,
        }

        if pid is not None:
            pobject['id'] = pid

        columns = ', '.join(pobject.keys())
        placeholders = ', '.join('?' * len(pobject))
        sql = 'INSERT INTO {} ({}) VALUES ({})'.format(tablename,
                                                       columns,
                                                       placeholders)
        cursor.execute(sql, list(pobject.values()))

        return cursor.lastrowid

    def addprimary(self, config, objauth, hierarchy='o'):

        c = self._conn.cursor()
        return self.addprimary_raw(c, 'pobjects', config, objauth, hierarchy)

    def addtertiary(self, tokid, pkcs11_object):
        tobject = {
            'tokid': tokid,
            'attrs': yaml.safe_dump(dict(pkcs11_object), canonical=True),
        }

        columns = ', '.join(tobject.keys())
        placeholders = ', '.join('?' * len(tobject))
        sql = 'INSERT INTO tobjects ({}) VALUES ({})'.format(columns,
                                                             placeholders)
        c = self._conn.cursor()
        c.execute(sql, list(tobject.values()))
        return c.lastrowid

    @staticmethod
    def _updatetertiary(db, tid, attrs):
        c = db.cursor()
        attrs = yaml.safe_dump(attrs, canonical=True)
        values = [attrs, tid]

        sql = 'UPDATE tobjects SET attrs=? WHERE id=?'
        c.execute(sql, values)

    def updatetertiary(self, tid, attrs):

        self._updatetertiary(self._conn, tid, attrs)

    def updatepin(self, is_so, token, sealauth, sealpriv, sealpub=None):

        tokid = token['id']

        c = self._conn.cursor()

        # TABLE sealobjects UPDATE
        # [user|so]priv TEXT NOT NULL,
        # [user|so]pub TEXT NOT NULL,
        # [user|so]authsalt TEXT NOT NULL,

        if sealpub:
            sql = 'UPDATE sealobjects SET {}authsalt=?, {}priv=?, {}pub=? WHERE id=?;'.format(
                * ['so' if is_so else 'user'] * 3)
            c.execute(sql, (sealauth['salt'], Db._blobify(sealpriv),
                            Db._blobify(sealpub), tokid))
        else:
            sql = 'UPDATE sealobjects SET {}authsalt=?, {}priv=? WHERE id=?;'.format(
                * ['so' if is_so else 'user'] * 2)
            c.execute(sql, (sealauth['salt'], Db._blobify(sealpriv), tokid))

    def commit(self):
        self._conn.commit()

    def __exit__(self, exc_type, exc_value, traceback):
        if (self._conn):
            self._conn.commit()
            self._conn.close()

    def delete(self):
        try:
            os.remove(self._path)
        except OSError:
            pass

    def backup(self):
        con = self._conn
        dbpath = self._path + ".bak"
        if os.path.exists(dbpath):
            raise RuntimeError("Backup DB exists at {} not overwriting. "
                "Refusing to run".format(dbpath))
        bck = sqlite3.connect(dbpath)
        con.backup(bck)
        return (bck, dbpath)

    def _update_on_2(self, dbbakcon):
        '''
        Between version 1 and 2 of the DB the following changes need to be made:
            The existing rows:
              - userpub BLOB NOT NULL,
              - userpriv BLOB NOT NULL,
              - userauthsalt TEXT NOT NULL,
            All have the "NOT NULL" constarint removed, like so:
                userpub BLOB,
                userpriv BLOB,
                userauthsalt TEXT
        So we need to create a new table with this constraint removed,
        copy the data and move the table back
        '''

        # Create a new table to copy data to that has the constraints removed
        s = textwrap.dedent('''
            CREATE TABLE sealobjects_new2(
                id INTEGER PRIMARY KEY,
                tokid INTEGER NOT NULL,
                userpub BLOB,
                userpriv BLOB,
                userauthsalt TEXT,
                sopub BLOB NOT NULL,
                sopriv BLOB NOT NULL,
                soauthsalt TEXT NOT NULL,
                FOREIGN KEY (tokid) REFERENCES tokens(id) ON DELETE CASCADE
            );
            ''')
        dbbakcon.execute(s)

        # copy the data
        s = textwrap.dedent('''
            INSERT INTO sealobjects_new2
            SELECT * FROM sealobjects;
        ''')
        dbbakcon.execute(s)

        # Drop the old table
        s = 'DROP TABLE sealobjects;'
        dbbakcon.execute(s)

        # Rename the new table to the correct table name
        s = 'ALTER TABLE sealobjects_new2 RENAME TO sealobjects;'
        dbbakcon.execute(s)

        # Add the triggers
        sql = [
            textwrap.dedent('''
                CREATE TRIGGER limit_tokens
                BEFORE INSERT ON tokens
                BEGIN
                    SELECT CASE WHEN
                        (SELECT COUNT (*) FROM tokens) >= 255
                    THEN
                        RAISE(FAIL, "Maximum token count of 255 reached.")
                    END;
                END;
            '''),
            textwrap.dedent('''
                CREATE TRIGGER limit_tobjects
                BEFORE INSERT ON tobjects
                BEGIN
                    SELECT CASE WHEN
                        (SELECT COUNT (*) FROM tobjects) >= 16777215
                    THEN
                        RAISE(FAIL, "Maximum object count of 16777215 reached.")
                    END;
                END;
            ''')
        ]
        for s in sql:
            dbbakcon.execute(s)

    def _update_on_3(self, dbbakcon):
        dbbakcon.execute('DROP TRIGGER limit_tobjects;')

    def _update_on_4(self, dbbakcon):
        '''
        Between version 3 and 1 of the DB the following changes need to be made:
        Table pobjects:
          - column handle of type blob was changes to config of type string

        The YAML config has the handle of the ESYS_TR blob as a hex string.
        So to perform the upgrade, the code needs to create a new db and copy
        everything over and generate the new config YAML as:
        ----
        persistent: true
        esys-tr: bytes.hex(handle)
        '''

        # Create a new table to copy data to
        s = textwrap.dedent('''
            CREATE TABLE pobjects2 (
                id INTEGER PRIMARY KEY,
                hierarchy TEXT NOT NULL,
                config TEXT NOT NULL,
                objauth TEXT NOT NULL
            );
            ''')
        dbbakcon.execute(s)

        c = dbbakcon.cursor()

        c.execute('SELECT * from pobjects')
        old_pobjects = c.fetchall()

        # copy the data and take the old handle (ESYS_TR Blob)
        # and convert to a hexstring and store in the config.
        # add the pobject to the new table keep the rest of the values.
        # Note: All migrating pobjects are persistent so mark as such.
        for pobj in old_pobjects:
            pid = pobj['id']
            blob = pobj['handle']
            hexblob = bytes.hex(blob)

            config = {
                'transient': False,
                'esys-tr': hexblob
            }

            objauth = pobj['objauth']
            hierarchy = pobj['hierarchy']
            self.addprimary_raw(c, 'pobjects2', config, objauth,
                                hierarchy, pid)

        # Drop the old table
        s = 'DROP TABLE pobjects;'
        dbbakcon.execute(s)

        # Rename the new table to the correct table name
        s = 'ALTER TABLE pobjects2 RENAME TO pobjects;'
        dbbakcon.execute(s)

    def _update_on_5(self, dbbakcon):
        '''
        Between version 4 and 5 of the DB the following changes need to be made:

        Table tobjects:

        The YAML attributes need to include CKM_AES_CBC_PAD and CKM_AES_CTR in the
        CKM_ALLOWED_MECHANISMS list.
        '''

        c = dbbakcon.cursor()

        c.execute('SELECT * from tobjects')
        tobjs = c.fetchall()

        for t in tobjs:
            attrs = yaml.safe_load(io.StringIO(t['attrs']))

            # IF the object is definitely a SECRET KEY of AES and has
            # CKM_AES_CBC_PAD AND CKM_AES_CTR in allowed mechanisms, skip it.
            if CKA_CLASS not in attrs or \
                attrs[CKA_CLASS] != CKO_SECRET_KEY or \
                CKA_KEY_TYPE not in attrs or \
                attrs[CKA_KEY_TYPE] != CKK_AES or \
                CKA_ALLOWED_MECHANISMS not in attrs \
                (CKM_AES_CBC_PAD in attrs[CKA_ALLOWED_MECHANISMS] \
                 and \
                 CKM_AES_CTR in attrs[CKA_ALLOWED_MECHANISMS]):
                continue

            # Is an AES KEY and needs CKM_AES_CBC_PAD
            if not CKM_AES_CBC_PAD in attrs[CKA_ALLOWED_MECHANISMS]:
                attrs[CKA_ALLOWED_MECHANISMS].append(CKM_AES_CBC_PAD)

            # Is an AES KEY and needs CKM_AES_CBC_PAD
            # This is a BUG that we add CBC_PAD twice, but we can't fix this
            # here and will have to do it on bump to version 6.
            if not CKM_AES_CTR in attrs[CKA_ALLOWED_MECHANISMS]:
                attrs[CKA_ALLOWED_MECHANISMS].append(CKM_AES_CBC_PAD)

            Db._updatetertiary(dbbakcon, t['id'], attrs)

    def _update_on_6(self, dbbakcon):
        '''
        Between version 5 and 6 of the DB the following changes need to be made:

        Table tobjects:

        The YAML attributes need to include CKM_ECDSA_SHA256 and CKM_ECDSA_SHA384,
        CKM_ECDSA_SHA512 in the CKM_ALLOWED_MECHANISMS list.
        '''

        c = dbbakcon.cursor()

        c.execute('SELECT * from tobjects')
        tobjs = c.fetchall()

        algs_to_add = set([ CKM_ECDSA_SHA256, CKM_ECDSA_SHA384, CKM_ECDSA_SHA512])

        for t in tobjs:
            attrs = yaml.safe_load(io.StringIO(t['attrs']))

            # Fix the duplicate add of CBC_PAD from version 4 -> 5 upgrade and
            # replace one with CBC_CTR
            if attrs[CKA_CLASS] == CKO_SECRET_KEY and \
                attrs[CKA_KEY_TYPE] == CKK_AES:
                    deduped_attrs = set(attrs[CKA_ALLOWED_MECHANISMS])
                    deduped_attrs.add(CKM_AES_CTR)
                    attrs[CKA_ALLOWED_MECHANISMS] = list(deduped_attrs)
                    Db._updatetertiary(dbbakcon, t['id'], attrs)
                    continue

            # IF the object is definitely a PRIVATE_KEY of EC and needs
            # one or more of the missing ECDSA algorithms added
            if CKA_CLASS not in attrs or \
                attrs[CKA_CLASS] != CKO_PRIVATE_KEY or \
                CKA_KEY_TYPE not in attrs or \
                attrs[CKA_KEY_TYPE] != CKK_EC:
                    # It is EC Private Key
                    allowed_mechs = set(attrs[CKA_ALLOWED_MECHANISMS] if CKA_ALLOWED_MECHANISMS in attrs else [])
                    # Needs one or more algs? If not skip
                    if algs_to_add.issubset(allowed_mechs):
                        continue

                    attrs[CKA_ALLOWED_MECHANISMS] = list(allowed_mechs | algs_to_add)
                    Db._updatetertiary(dbbakcon, t['id'], attrs)

    def _update_on_7(self, dbbakcon):
        '''
        Between version 6 and 7 of the DB the following changes need to be made:

        Table tobjects:

        Filter out possible 0 mechanisms from CKM_ALLOWED_MECHANISMS list due to
        previous bugs in db update code.
        '''

        c = dbbakcon.cursor()

        c.execute('SELECT * from tobjects')
        tobjs = c.fetchall()

        for t in tobjs:
            attrs = yaml.safe_load(io.StringIO(t['attrs']))

            # Fix the duplicate add of CBC_PAD from version 4 -> 5 upgrade and
            # replace one with CBC_CTR
            if CKA_KEY_TYPE not in attrs:
                continue

            cka_class = attrs[CKA_CLASS]
            cka_key_type = attrs[CKA_KEY_TYPE]
            if cka_class in [ CKO_SECRET_KEY, CKO_PRIVATE_KEY ] and \
                cka_key_type in [ CKK_AES, CKK_EC ] and \
                0 in attrs[CKA_ALLOWED_MECHANISMS]:
                    deduped_attrs = set(attrs[CKA_ALLOWED_MECHANISMS])
                    # we checked that 0 was in attrs.
                    deduped_attrs.remove(0)
                    attrs[CKA_ALLOWED_MECHANISMS] = list(deduped_attrs)
                    Db._updatetertiary(dbbakcon, t['id'], attrs)

    def _update_on_8(self, dbbakcon):
        '''
        Between version 7 and 8 of the DB the following changes need to be made:

        Table tobjects:

        The YAML attributes for a int sequence has to be a yaml sequence.
        '''
        import ctypes
        import socket
        long_size = ctypes.sizeof(ctypes.c_ulong)

        c = dbbakcon.cursor()

        c.execute('SELECT * from tobjects')
        tobjs = c.fetchall()

        for t in tobjs:
            attrs = yaml.safe_load(io.StringIO(t['attrs']))
            for attr in attrs:
                # The allowed mechanism attribute is a buffer of hexadecimal
                # written as a string instead of being a sequence of int
                if attr == CKA_ALLOWED_MECHANISMS and \
                    isinstance(attrs[attr], str):
                        list_hexa = [socket.ntohl(int(attrs[attr][i:i+long_size], 16)) for i in range(0, len(attrs[attr]), long_size)]
                        list_int = [int(x,16) for x in list_hexa]
                        attrs[attr] = list_int

            Db._updatetertiary(dbbakcon, t['id'], attrs)

    def update_db(self, old_version, new_version=VERSION):

        # were doing the update, so make a backup to manipulate
        (dbbakcon, dbbakpath) = self.backup()

        dbbakcon.row_factory = sqlite3.Row

        try:
            for x in range(old_version, new_version):
                x = x + 1
                getattr(self, '_update_on_{}'.format(x))(dbbakcon)

            sql = textwrap.dedent('''
                    REPLACE INTO schema (id, schema_version) VALUES (1, {version});
                '''.format(version=new_version))
            dbbakcon.execute(sql)
        except Exception as e:
            # Close the connection to backup
            dbbakcon.commit()
            dbbakcon.close()

            # unlink the backup
            os.unlink(dbbakpath)

            raise e

        # Close the connections
        self._conn.commit()
        self._conn.close()

        dbbakcon.commit()
        dbbakcon.close()

        # move old db to ".old" suffix
        olddbpath = self._path + ".old"
        os.rename(self._path, olddbpath)

        # move the backup to the normal dbpath
        os.rename(dbbakpath, self._path)

        # re-establish a connection
        self._conn = sqlite3.connect(self._path)
        self._conn.row_factory = sqlite3.Row

    def _get_version(self):
        c = self._conn.cursor()
        try:
            c.execute('select schema_version from schema')
            return c.fetchone()[0]
        except sqlite3.OperationalError:
            return 0

    def db_init_new(self):

        c = self._conn.cursor()

        sql = [
            textwrap.dedent('''
            CREATE TABLE tokens(
                id INTEGER PRIMARY KEY,
                pid INTEGER NOT NULL,
                label TEXT UNIQUE,
                config TEXT NOT NULL,
                FOREIGN KEY (pid) REFERENCES pobjects(id) ON DELETE CASCADE
            );
            '''),
            textwrap.dedent('''
            CREATE TABLE sealobjects(
                id INTEGER PRIMARY KEY,
                tokid INTEGER NOT NULL,
                userpub BLOB,
                userpriv BLOB,
                userauthsalt TEXT,
                sopub BLOB NOT NULL,
                sopriv BLOB NOT NULL,
                soauthsalt TEXT NOT NULL,
                FOREIGN KEY (tokid) REFERENCES tokens(id) ON DELETE CASCADE
            );
            '''),
            textwrap.dedent('''
            CREATE TABLE pobjects(
                id INTEGER PRIMARY KEY,
                hierarchy TEXT NOT NULL,
                config TEXT NOT NULL,
                objauth TEXT NOT NULL
            );
            '''),
            textwrap.dedent('''
            CREATE TABLE tobjects(
                id INTEGER PRIMARY KEY,
                tokid INTEGER NOT NULL,
                attrs TEXT NOT NULL,
                FOREIGN KEY (tokid) REFERENCES tokens(id) ON DELETE CASCADE
            );
            '''),
            textwrap.dedent('''
            CREATE TABLE schema(
                id INTEGER PRIMARY KEY,
                schema_version INTEGER NOT NULL
            );
            '''),
            textwrap.dedent('''
                CREATE TRIGGER limit_tokens
                BEFORE INSERT ON tokens
                BEGIN
                    SELECT CASE WHEN
                        (SELECT COUNT (*) FROM tokens) >= 255
                    THEN
                        RAISE(FAIL, "Maximum token count of 255 reached.")
                    END;
                END;
            '''),
            textwrap.dedent('''
                CREATE TRIGGER limit_tobjects
                BEFORE INSERT ON tobjects
                BEGIN
                    SELECT CASE WHEN
                        (SELECT COUNT (*) FROM tobjects) >= 16777215
                    THEN
                        RAISE(FAIL, "Maximum object count of 16777215 reached.")
                    END;
                END;
            '''),
            textwrap.dedent('''
                REPLACE INTO schema (id, schema_version) VALUES (1, {version});
            '''.format(version=VERSION))
        ]

        for s in sql:
            c.execute(s)

    # TODO collapse object tables into one, since they are common besides type.
    # move sealobject metadata into token metadata table.
    #
    # Object types:
    # soseal
    # userseal
    # wrapping
    # secondary
    # tertiary
    #
    # NOTE: Update the DB Schema Version at the bottom if the db format changes!
    def _do_create(self):

        # perform an update if we need to
        try:
            old_version = self._get_version()

            if old_version == 0:
                self.db_init_new()
                self.version = VERSION
                self.VERSION = VERSION
                return False
            elif VERSION == old_version:
                self.version = old_version
                self.VERSION = old_version
                return False
            elif old_version > VERSION:
                raise RuntimeError("DB Version exceeds library version:"
                 " {} > {}".format(old_version, VERSION))
            else:
                self.version = old_version
                self.update_db(old_version, VERSION)
                self.VERSION = self._get_version()
                return True

        except Exception as e:
            sys.stderr.write('DB Upgrade failed: "{}"\n'.format(e))
            raise e

    def _create(self):

        # create a lock from the db name plush .lock suffix
        lockpath = self._path+".lock"
        holds_lock = False
        fd = os.open(lockpath, os.O_CREAT|os.O_RDWR)
        try:
            fcntl.flock(fd, fcntl.LOCK_EX)
            holds_lock = True
            self._do_create()
        finally:
            # we always want unlink to occur
            os.unlink(lockpath)
            if holds_lock:
                fcntl.flock(fd, fcntl.LOCK_UN)
            os.close(fd)



# SPDX-License-Identifier: BSD-2-Clause
# python stdlib dependencies
import binascii
import io
import sys

# External dependencies
import yaml

# local imports
from .command import Command
from .command import commandlet
from .db import Db
from .utils import check_pss_signature
from .utils import TemporaryDirectory
from .utils import hash_pass
from .utils import rand_hex_str
from .utils import AESAuthUnwrapper
from .utils import load_sealobject
from .utils import str2bool
from .utils import pkcs11_cko_to_str
from .utils import pkcs11_ckk_to_str
from .utils import get_pobject
from .tpm2 import Tpm2

from .pkcs11t import *  # noqa

@commandlet("rmtoken")
class RmTokenCommand(Command):
    '''
    Removes a token from a tpm2-pkcs11 store
    '''

    # adhere to an interface
    # pylint: disable=no-self-use
    def generate_options(self, group_parser):
        group_parser.add_argument(
            '--label', help='The profile label to remove.\n', required=True)

    def __call__(self, args):

        path = args['path']
        label = args['label']

        with Db(path) as db:
            token = db.gettoken(label)
            db.rmtoken(token['label'])


@commandlet("verify")
class VerifyCommand(Command):
    '''
    Verifies the userpin and/or sopin for a given profile.
    '''

    # adhere to an interface
    # pylint: disable=no-self-use
    def generate_options(self, group_parser):
        group_parser.add_argument(
            '--sopin',
            help='The Administrator pin. This pin is used for object recovery.\n'
        )
        group_parser.add_argument(
            '--userpin',
            help='The user pin. This pin is used for authentication for object use.\n'
        )
        group_parser.add_argument(
            '--label', help='The label to verify.\n', required=True)
        group_parser.add_argument(
            '--hierarchy-auth',
            help='The authorization password for the owner hierarchy when using a token with a transient primary object\n',
            default="")
    @staticmethod
    def verify(db, args):

        label = args['label']

        token = db.gettoken(label)
        if token is None:
            sys.exit('No token labeled "%s"' % label)

        token_config = yaml.safe_load(io.StringIO(token['config']))

        sopin = args['sopin']
        userpin = args['userpin']
        hierarchyauth = args['hierarchy_auth']

        if userpin is None and sopin is None:
            # Use empty PIN if the token has an empty user PIN
            if token_config.get('empty-user-pin'):
                userpin = ''
            else:
                sys.exit('error: at least one of the arguments --sopin --userpin is required')

        verify_output = {}
        verify_output['label'] = label

        pobj = db.getprimary(token['pid'])
        sealobj = db.getsealobject(token['id'])

        wrappingkeyauth = None

        verify_output['config'] = token_config

        verify_output['pin'] = {}

        with TemporaryDirectory() as d:
            tpm2 = Tpm2(d)

            pobjauth = pobj['objauth']
            pobj_handle = get_pobject(pobj, tpm2, hierarchyauth, d)

            if sopin != None:

                sosealctx = tpm2.load(pobj_handle, pobjauth,
                                      sealobj['sopriv'], sealobj['sopub'])

                # Unseal the wrapping key auth
                sosealauthsalt = sealobj['soauthsalt']

                sosealauth = hash_pass(sopin, salt=sosealauthsalt)

                wrappingkeyauth = tpm2.unseal(sosealctx, sosealauth['hash'])

                verify_output['pin']['so'] = {'seal-auth' : sosealauth['hash'] }

            if userpin != None:

                usersealctx = tpm2.load(pobj_handle, pobjauth,
                                        sealobj['userpriv'],
                                        sealobj['userpub'])

                # Unseal the wrapping key auth
                usersealauthsalt = sealobj['userauthsalt']

                usersealauth = hash_pass(userpin, salt=usersealauthsalt)

                wrappingkeyauth = tpm2.unseal(usersealctx,
                                              usersealauth['hash'])

                verify_output['pin']['user'] = {'seal-auth' : usersealauth['hash'] }

            verify_output['wrappingkey'] = {
                'hex' : bytes.hex(wrappingkeyauth),
            }
            if userpin != None:
                verify_output['wrappingkey']['auth'] = usersealauth['hash']
            if sopin != None:
                verify_output['wrappingkey']['soauth'] = sosealauth['hash']

            wrapper = AESAuthUnwrapper(wrappingkeyauth)

            tobjs = db.gettertiary(token['id'])

            verify_output['objects'] = []

            for tobj in tobjs:

                attrs = yaml.safe_load(tobj['attrs'])

                priv=None
                if CKA_TPM2_PRIV_BLOB in attrs:
                    priv = binascii.unhexlify(attrs[CKA_TPM2_PRIV_BLOB])

                pub = None
                if CKA_TPM2_PUB_BLOB in attrs:
                    pub = binascii.unhexlify(attrs[CKA_TPM2_PUB_BLOB])

                encauth = None
                if CKA_TPM2_OBJAUTH_ENC in attrs:
                    encauth = binascii.unhexlify(attrs[CKA_TPM2_OBJAUTH_ENC])

                tobjauth=None
                if encauth:
                    encauth=encauth.decode()
                    tpm2.load(pobj_handle, pobjauth, priv, pub)
                    tobjauth = wrapper.unwrap(encauth).decode()

                verify_output['objects'].append({
                    'id' : tobj['id'],
                    'auth' : tobjauth,
                    'encauth' : encauth
                })

        yaml_dump = yaml.safe_dump(verify_output, default_flow_style=False)
        print(yaml_dump)

    def __call__(self, args):
        with Db(args['path']) as db:
            VerifyCommand.verify(db, args)


@commandlet("addtoken")
class AddTokenCommand(Command):
    '''
    Adds an initialized token to a tpm2-pkcs11 store
    '''

    # adhere to an interface
    # pylint: disable=no-self-use
    def generate_options(self, group_parser):
        group_parser.add_argument(
            '--pid',
            type=int,
            help='The primary object id to associate with this token.\n',
            required=True),
        group_parser.add_argument(
            '--sopin',
            help='The Administrator pin. This pin is used for object recovery.\n',
            required=True),
        group_parser.add_argument(
            '--userpin',
            help='The user pin. This pin is used for authentication for object use.\n',
            required=True),
        group_parser.add_argument(
            '--label',
            help='A unique label to identify the profile in use, must be unique.\n',
            required=True)
        group_parser.add_argument(
            '--hierarchy-auth',
            help='The authorization password for the owner hierarchy when using a token with a transient primary object\n',
            default="")
    @staticmethod
    def do_token_init(db, path, args):

        userpin = args['userpin']
        sopin = args['sopin']
        label = args['label']
        pid = args['pid']
        hierarchyauth = args['hierarchy_auth']

        # Verify pid is in db
        pobject = db.getprimary(pid)
        if not pobject:
            raise RuntimeError('No primary object id: %u' % (pid))

        with TemporaryDirectory() as d:
            tpm2 = Tpm2(d)

            # generate a wrapping key to be sealed to the TPM.
            # AES 256 key is 32 bytes, or 64 hex chars.
            wrappingkey = rand_hex_str(64)

            # We generate one auth for the sosealobj and the usersealobj
            sosealauth = hash_pass(sopin)
            usersealauth = hash_pass(userpin)

            # Now we generate the two seal objects, using the sealauths as their
            # auth values and sealing the wrappingkey value to it.
            # soobject will be an AES key used for decrypting tertiary object
            # auth values.
            pobj_handle = get_pobject(pobject, tpm2, hierarchyauth, d)

            usersealpriv, usersealpub, _ = tpm2.create(
                pobj_handle,
                pobject['objauth'],
                usersealauth['hash'],
                seal=wrappingkey)
            sosealpriv, sosealpub, _ = tpm2.create(
                pobj_handle,
                pobject['objauth'],
                sosealauth['hash'],
                seal=wrappingkey)

            pss_sig_good = check_pss_signature(tpm2, pobj_handle, pobject['objauth'])

            # If this succeeds, we update the token table
            config = {
                'token-init': True,
                'pss-sigs-good' : pss_sig_good
            }
            if userpin == '':
                config['empty-user-pin'] = True
            tokid = db.addtoken(pobject['id'], config, label=label)

            # now we update the sealobject table with the tokid to seal objects mapping
            db.addsealobjects(tokid, usersealauth, usersealpriv, usersealpub,
                              sosealauth, sosealpriv, sosealpub)

    @staticmethod
    def do_token_noninit(db, args):

        pid = args['pid']
        hierarchyauth = args['hierarchy_auth']

        with TemporaryDirectory() as d:
            tpm2 = Tpm2(d)

            pobject = db.getprimary(pid)

            pobjauth = pobject['objauth']

            pobj_handle = get_pobject(pobject, tpm2, hierarchyauth, d)

            pss_sig_good = check_pss_signature(tpm2, pobj_handle, pobjauth)

            # If this succeeds, we update the token table
            config = {
                'token-init': False,
                'pss-sigs-good' : pss_sig_good
            }

            db.addtoken(pobject['id'], config)

    def __call__(self, args):

        path = args['path']
        do_token_init = not args.get('no_init', False)

        with Db(path) as db:

            if do_token_init:
                AddTokenCommand.do_token_init(db, path, args)
            else:
                AddTokenCommand.do_token_noninit(db, args)

@commandlet("addemptytoken")
class AddEmptyTokenCommand(AddTokenCommand):
    '''
    Adds an un-initialized token to a tpm2-pkcs11 store.
    '''

    def generate_options(self, group_parser):
        group_parser.add_argument(
            '--pid',
            type=int,
            help='The primary object id to associate with this token.\n',
            required=True),
        group_parser.add_argument(
            '--hierarchy-auth',
            help='The authorization password for the owner hierarchy when using a token with a transient primary object\n',
            default="")

    def __call__(self, args):
        args['no_init'] = True
        super(AddEmptyTokenCommand, self).__call__(args)


@commandlet("changepin")
class ChangePinCommand(Command):
    '''
    Changes the userpin and/or sopin for a given token.
    '''

    # adhere to an interface
    # pylint: disable=no-self-use
    def generate_options(self, group_parser):
        group_parser.add_argument(
            '--user',
            choices=['so', 'user'],
            default='user',
            help='Which pin to change. Defaults to "user".\n')
        group_parser.add_argument(
            '--old', type=str, help='The old pin.\n')
        group_parser.add_argument(
            '--new', type=str, help='The new pin.\n', required=True)
        group_parser.add_argument(
            '--label',
            type=str,
            help='The label of the token.\n',
            required=True)
        group_parser.add_argument(
            '--hierarchy-auth',
            help='The authorization password for the owner hierarchy when using a token with a transient primary object\n',
            default="")

    @staticmethod
    def changepin(db, tpm2, args):

        label = args['label']

        is_so = args['user'] == 'so'
        oldpin = args['old']
        newpin = args['new']
        hierarchyauth = args['hierarchy_auth']

        token = db.gettoken(label)
        token_config = yaml.safe_load(io.StringIO(token['config']))
        if oldpin is None:
            if is_so:
                sys.exit('error: the SO PIN is missing, argument --old is required')
            # Use empty PIN if the token has an empty user PIN
            if token_config.get('empty-user-pin'):
                oldpin = ''
            else:
                sys.exit('error: the user PIN is missing, argument --old is required')

        pobjectid = token['pid']
        pobject = db.getprimary(pobjectid)
        pobjauth = pobject['objauth']

        with TemporaryDirectory() as d:

            pobj_handle = get_pobject(pobject, tpm2, hierarchyauth, d)

            sealctx, sealauth = load_sealobject(token, db, tpm2, pobj_handle, pobjauth,
                                                      oldpin, is_so)

            newsealauth = hash_pass(newpin)

            # call tpm2_changeauth and get new private portion
            newsealpriv = tpm2.changeauth(pobj_handle, sealctx, sealauth,
                                          newsealauth['hash'])

        # update 'empty-user-pin' in a safe way: clear it before setting a
        # non-empty PIN and set it after setting an empty PIN
        if not is_so and newpin != '' and token_config.get('empty-user-pin'):
            del token_config['empty-user-pin']
            db.updateconfig(token, token_config)

        # update the database
        db.updatepin(is_so, token, newsealauth, newsealpriv)

        if not is_so and newpin == '' and not token_config.get('empty-user-pin'):
            token_config['empty-user-pin'] = True
            db.updateconfig(token, token_config)

    def __call__(self, args):

        path = args['path']

        with Db(path) as db:
            with TemporaryDirectory() as d:
                tpm2 = Tpm2(d)
                ChangePinCommand.changepin(db, tpm2, args)


@commandlet("initpin")
class InitPinCommand(Command):
    '''
    Resets the userpin given a sopin for a given token.
    '''

    # adhere to an interface
    # pylint: disable=no-self-use
    def generate_options(self, group_parser):
        group_parser.add_argument(
            '--sopin', type=str, help='The current sopin.\n', required=True)
        group_parser.add_argument(
            '--userpin', type=str, help='The new user pin.\n', required=True)
        group_parser.add_argument(
            '--label',
            type=str,
            help='The label of the token.\n',
            required=True)
        group_parser.add_argument(
            '--hierarchy-auth',
            help='The authorization password for the owner hierarchy when using a token with a transient primary object\n',
            default="")

    @staticmethod
    def initpin(db, tpm2, args):

        label = args['label']

        sopin = args['sopin']
        newpin = args['userpin']

        token = db.gettoken(label)
        token_config = yaml.safe_load(io.StringIO(token['config']))

        # load and unseal the data from the SO seal object
        pobjectid = token['pid']
        pobject = db.getprimary(pobjectid)
        pobjauth = pobject['objauth']
        hierarchyauth = args['hierarchy_auth']

        with TemporaryDirectory() as d:

            pobj_handle = get_pobject(pobject, tpm2, hierarchyauth, d)

            sealctx, sealauth = load_sealobject(token, db, tpm2, pobj_handle, pobjauth,
                                                  sopin, True)
            wrappingkeyauth = tpm2.unseal(sealctx, sealauth)

            # call tpm2_create and create a new sealobject protected by the seal auth and sealing
            #    the wrapping key auth value
            newsealauth = hash_pass(newpin)

            newsealpriv, newsealpub, _ = tpm2.create(
                pobj_handle,
                pobjauth,
                newsealauth['hash'],
                seal=wrappingkeyauth)

        # update 'empty-user-pin' in a safe way
        if newpin != '' and token_config.get('empty-user-pin'):
            del token_config['empty-user-pin']
            db.updateconfig(token, token_config)

        # update the database
        db.updatepin(False, token, newsealauth, newsealpriv, newsealpub)

        if newpin == '' and not token_config.get('empty-user-pin'):
            token_config['empty-user-pin'] = True
            db.updateconfig(token, token_config)

    def __call__(self, args):

        path = args['path']

        with Db(path) as db:
            with TemporaryDirectory() as d:
                tpm2 = Tpm2(d)
                InitPinCommand.initpin(db, tpm2, args)

@staticmethod
def _empty_validator(s):
    return s


def _forbid_set_empty_user_pin(_):
    raise RuntimeError("'empty-user-pin' can only be set with changepin or initpin")


@commandlet("config")
class ConfigCommand(Command):
    '''
    Manipulates and retrieves token configuration data.
    '''
    _keys = {
        'token-init' : str2bool,
        'log-level'  : _empty_validator.__func__,
        'tcti'       : _empty_validator.__func__,
        'empty-user-pin': _forbid_set_empty_user_pin,
    }

    # adhere to an interface
    # pylint: disable=no-self-use
    def generate_options(self, group_parser):
        group_parser.add_argument(
            '--key', type=str, help='The key to set, valid keys are: %s.\n' % self._keys.keys())
        group_parser.add_argument(
            '--value', type=str, help='The value for the key.\n')
        group_parser.add_argument(
            '--label',
            type=str,
            help='The label of the token.\n',
            required=True)

    @classmethod
    def get_validator_for_key(cls, key):
        return cls._keys[key]

    @classmethod
    def config(cls, db, args):

        label = args['label']

        key = args['key']
        value = args['value']

        token = db.gettoken(label)

        token_config = yaml.safe_load(io.StringIO(token['config']))
        if not key and not value:
            yaml_tok_cconf = yaml.safe_dump(token_config, default_flow_style=False)
            print(yaml_tok_cconf)
            sys.exit(0)

        if not key and value:
            sys.exit("Cannot specify --value without a key")

        if key == 'log-level':
            print('WARN --key="log-level is deprecated', file=sys.stderr)

        # key has to be set here based on above logical check
        # throws an error if the key isn't known to the system
        validator = cls.get_validator_for_key(key)

        # no value, just key. Print the current value for key is set or empty if not set
        if not value:
            print("%s=%s" % (key, str(token_config[key] if key in token_config else "")))
            sys.exit(0)

        # bitbucket log-level sets
        if key == 'log-level':
            print('WARN --key="log-level is ignored', file=sys.stderr)
            return

        v = validator(value)
        token_config[key] = v

        # update the database
        db.updateconfig(token, token_config)

    def __call__(self, args):

        path = args['path']

        with Db(path) as db:
            ConfigCommand.config(db, args)

@commandlet("listprimaries")
class ListPrimaryCommand(Command):
    '''
    Lists primary objects in  a specified store.
    '''

    # adhere to an interface
    # pylint: disable=no-self-use
    def generate_options(self, group_parser):
        pass

    @staticmethod
    def list(db):
        output=[]
        primaries = db.getprimaries()
        for p in primaries:
            details = {'id': p['id']}
            details['config'] = yaml.safe_load(p['config'])
            output.append(details)

        if len(output):
            print(yaml.safe_dump(output, default_flow_style=False))

    def __call__(self, args):

        path = args['path']

        with Db(path) as db:
            ListPrimaryCommand.list(db)

@commandlet("listtokens")
class ListTokenCommand(Command):
    '''
    Lists tokens in  a specified store.
    '''

    # adhere to an interface
    # pylint: disable=no-self-use
    def generate_options(self, group_parser):
        group_parser.add_argument(
            '--pid',
            type=int,
            help='The primary object id to associate with this token.\n',
            required=True),

    @staticmethod
    def list(db, args):
        output = []
        tokens = db.gettokens(args['pid'])

        for t in tokens:
            output.append({
                'id': t['id'],
                'label': t['label']
            })

        if len(output):
            print(yaml.safe_dump(output, default_flow_style=False))

    def __call__(self, args):

        path = args['path']

        with Db(path) as db:
            ListTokenCommand.list(db, args)

@commandlet("listobjects")
class ListObjectsCommand(Command):
    '''
    Lists Objects (keys, certificates, etc.) associated with a token.
    '''

    # adhere to an interface
    # pylint: disable=no-self-use
    def generate_options(self, group_parser):
        group_parser.add_argument(
            '--label',
            type=str,
            help='The label of the token.\n',
            required=True)

    @staticmethod
    def list(db, args):
        output = []
        token = db.gettoken(args['label'])
        objects = db.getobjects(token['id'])
        for o in objects:
            y = yaml.safe_load(o['attrs'])
            d = {
                'id': o['id'],
                'CKA_LABEL' : binascii.unhexlify(y[CKA_LABEL]).decode(),
                'CKA_CLASS' : pkcs11_cko_to_str(y[CKA_CLASS]),
            }

            if CKA_ID in y:
                d['CKA_ID'] = y[CKA_ID],

            if CKA_KEY_TYPE in y:
                d['CKA_KEY_TYPE'] = pkcs11_ckk_to_str(y[CKA_KEY_TYPE])

            output.append(d)

        if len(output):
            print(yaml.safe_dump(output, default_flow_style=False))

    def __call__(self, args):

        path = args['path']

        with Db(path) as db:
            ListObjectsCommand.list(db, args)



# SPDX-License-Identifier: BSD-2-Clause
# python stdlib dependencies
import binascii
import io
import os
import struct
import sys
import yaml

from tempfile import mkstemp

from tpm2_pytss.tsskey import TSSPrivKey

# local imports
from .command import Command
from .command import commandlet
from .db import Db
from .objects import PKCS11ObjectFactory as PKCS11ObjectFactory
from .objects import PKCS11X509
from .utils import AESAuthUnwrapper
from .utils import TemporaryDirectory
from .utils import hash_pass
from .utils import rand_hex_str
from .utils import pemcert_to_attrs
from .utils import str2bool
from .utils import str2bytes
from .utils import get_pobject
from .utils import dump_blobs
from .utils import dump_tsspem
from .utils import dump_pubpem

from .tpm2 import Tpm2

from .pkcs11t import *  # noqa

class NewKeyCommandBase(Command):
    '''
    creates a key to a token within a tpm2-pkcs11 store.
    '''

    def generate_options(self, group_parser):
        group_parser.add_argument(
            '--label',
            help='The tokens label to import the key too.\n',
            required=True)
        group_parser.add_argument(
            '--key-label',
            help='The label of the key imported. Defaults to an integer value.\n')
        group_parser.add_argument(
            '--id',
            help='The key id. Defaults to a random 8 bytes of hex.\n',
            default=binascii.hexlify(os.urandom(8)).decode())
        group_parser.add_argument(
            '--attr-always-authenticate',
            action='store_true',
            help='Sets the CKA_ALWAYS_AUTHENTICATE attribute to CK_TRUE.\n')
        group_parser.add_argument(
            '--hierarchy-auth',
            help='The hierarchyauth, required for transient pobjects.\n',
            default='')
        pinopts = group_parser.add_mutually_exclusive_group()
        pinopts.add_argument('--sopin', help='The Administrator pin.\n'),
        pinopts.add_argument('--userpin', help='The User pin.\n'),

    # Implemented by derived class
    def new_key_create(self, pobj, objauth, hierarchyauth, tpm2, alg, privkey, passin, d):
        raise NotImplementedError('Implement: new_key')

    def new_key_init(self, label, sopin, userpin, hierarchyauth, pobj, sealobjects, tpm2, d):

        pobj_handle = get_pobject(pobj, tpm2, hierarchyauth, d)

        # Get the primary object encrypted auth value and sokey information
        # to decode it. Based on the incoming pin
        is_so = sopin != None
        pin = sopin if is_so else userpin

        pubkey = '%spub' % ('so' if is_so else 'user')
        privkey = '%spriv' % ('so' if is_so else 'user')
        saltkey = '%sauthsalt' % ('so' if is_so else 'user')

        sealpub = sealobjects[pubkey]
        sealpriv = sealobjects[privkey]
        sealsalt = sealobjects[saltkey]

        sealctx = tpm2.load(pobj_handle, pobj['objauth'], sealpriv, sealpub)

        sealauth = hash_pass(pin, salt=sealsalt)['hash']

        wrappingkey = tpm2.unseal(sealctx, sealauth)

        wrapper = AESAuthUnwrapper(wrappingkey)

        #create an auth value for the tertiary object.
        objauth = rand_hex_str()

        encobjauth = wrapper.wrap(str2bytes(objauth))

        return (encobjauth, objauth)

    @staticmethod
    def new_key_save(alg, keylabel, tid, label, privblob, pubblob,
                     tertiarypubdata, encobjauth, db, tpm2, extra_privattrs=None, extra_pubattrs=None,
                     override_keylen=None):
        token = db.gettoken(label)

        #
        # Cache the objects attributes from the public structure and other sources
        # and populate the db with the data. This allows use of the public data
        # without needed to load any objects which requires a pin to do.
        #
        y = yaml.safe_load(tertiarypubdata)

        initial_pubattrs = {}
        initial_privattrs = {}

        # add the id
        initial_privattrs.update({CKA_ID: binascii.hexlify(tid.encode()).decode()})
        initial_pubattrs.update({CKA_ID: binascii.hexlify(tid.encode()).decode()})

        # Add keylabel for ALL objects if set
        if keylabel is not None:
            initial_privattrs.update({
                CKA_LABEL: binascii.hexlify(keylabel.encode()).decode()
            })
            initial_pubattrs.update({
                CKA_LABEL: binascii.hexlify(keylabel.encode()).decode()
            })

        # add additional attrs
        if extra_privattrs:
            initial_privattrs.update(extra_privattrs)

        if initial_pubattrs and extra_pubattrs:
            initial_pubattrs.update(extra_pubattrs)

        objects = PKCS11ObjectFactory(y, tpm2, encobjauth, initial_pubattrs, initial_privattrs,
                      tpm_pub=pubblob, tpm_priv=privblob, override_keylen=override_keylen)

        # Store private to database
        db.addtertiary(token['id'], objects['private'])

        # if it's asymmetric, add a public object too
        if 'public' in objects and objects['public'] is not None:
            db.addtertiary(token['id'], objects['public'])

        return objects

    @staticmethod
    def output(objects, action):
        d = {
            'action' : action,
        }

        for k, v in objects.items():
            if v is not None:
                d[k] = { 'CKA_ID' : objects[k][CKA_ID] }

        yaml.safe_dump(d, sys.stdout, default_flow_style=False)

    def __call__(self, args):
        path = args['path']

        with Db(path) as db:

            with TemporaryDirectory() as d:
                tpm2 = Tpm2(d)

                label = args['label']
                sopin = args['sopin']
                userpin = args['userpin']
                alg = args['algorithm'] if 'algorithm' in args else None
                key_label = args['key_label']
                tid = args['id']
                hierarchyauth = args['hierarchy_auth']
                passin = args['passin'] if 'passin' in args else None

                privkey = None
                try:
                    privkey = args['privkey']
                except KeyError:
                    privkey = None

                token = db.gettoken(label)
                pobjectid = token['pid']
                pobj = db.getprimary(pobjectid)

                if userpin is None and sopin is None:
                    # Use empty PIN if the token has an empty user PIN
                    token_config = yaml.safe_load(io.StringIO(token['config']))
                    if token_config.get('empty-user-pin'):
                        userpin = ''
                    else:
                        sys.exit('error: at least one of the arguments --sopin --userpin is required')

                sealobjects = db.getsealobject(token['id'])

                encobjauth, objauth = self.new_key_init(
                    label, sopin, userpin, hierarchyauth,
                    pobj, sealobjects, tpm2, d)

                tertiarypriv, tertiarypub, tertiarypubdata = self.new_key_create(
                    pobj, objauth, hierarchyauth, tpm2, alg, privkey, passin, d)

                # handle options that can add additional attributes
                always_auth = args['attr_always_authenticate']
                priv_attrs = {CKA_ALWAYS_AUTHENTICATE : always_auth}

                override_keylen = getattr(self, '_override_keylen', None)

                return NewKeyCommandBase.new_key_save(
                    alg, key_label, tid, label, tertiarypriv, tertiarypub,
                    tertiarypubdata, encobjauth, db, tpm2, extra_privattrs=priv_attrs,
                    override_keylen=override_keylen)


@commandlet("import")
class ImportCommand(NewKeyCommandBase):
    '''
    Imports a rsa key to a token within a tpm2-pkcs11 store.
    '''

    # adhere to an interface
    # pylint: disable=no-self-use
    def generate_options(self, group_parser):
        super(ImportCommand, self).generate_options(group_parser)
        group_parser.add_argument(
            '--privkey',
            help='Full path of the private key to be imported.\n',
            required=True)
        group_parser.add_argument(
            '--algorithm',
            help='The type of the key.\n',
            choices=['rsa', 'ecc',
                     'hmac', # Generic hmac, actually imported as keyedhash scheme null with proper attrs.
                     'hmac:sha1', 'hmac:sha256', 'hmac:sha384', 'hmac:sha512' ],
            required=False)
        group_parser.add_argument(
            '--passin',
            help='Password of the input private key file like OpenSSL (example: "pass:secret").\n',
            required=False)

    # Imports a new key
    def new_key_create(self, pobj, objauth, hierarchyauth, tpm2, alg, privkey, passin, d):

        pobj_handle = get_pobject(pobj, tpm2, hierarchyauth, d)

        # A plain HMAC key is just a keyedhash object with userwithauth and sign set.
        # Seal objects don't have sign (ie default tpm2_import -G keyedhash behavior.
        objattrs=None
        if alg == 'hmac':
            objattrs = 'userwithauth|sign|decrypt'
            alg = 'keyedhash'

        tertiarypriv, tertiarypub, tertiarypubdata = tpm2.importkey(
            pobj_handle, pobj['objauth'], objauth, privkey=privkey, alg=alg, passin=passin, objattrs=objattrs)

        # We have no way of knowing the keylength of an hmac key
        if alg and alg.startswith('hmac') or alg == 'keyedhash':
            self._override_keylen = os.path.getsize(privkey)

        return (tertiarypriv, tertiarypub, tertiarypubdata)

    def __call__(self, args):
        objects = super(ImportCommand, self).__call__(args)
        NewKeyCommandBase.output(objects, 'import')

@commandlet("addkey")
class AddKeyCommand(NewKeyCommandBase):
    '''
    Adds a key to a token within a tpm2-pkcs11 store.
    '''

    # adhere to an interface
    # pylint: disable=no-self-use
    def generate_options(self, group_parser):
        super(AddKeyCommand, self).generate_options(group_parser)
        group_parser.add_argument(
            '--algorithm',
            help='The type of the key.\n',
            choices=Tpm2.ALGS,
            required=True)

    # Creates a new key
    def new_key_create(self, pobj, objauth, hierarchyauth, tpm2, alg, privkey, passin, d):

        pobj_handle = get_pobject(pobj, tpm2, hierarchyauth, d)

        tertiarypriv, tertiarypub, tertiarypubdata = tpm2.create(
            pobj_handle, pobj['objauth'], objauth, alg=alg)

        return (tertiarypriv, tertiarypub, tertiarypubdata)

    def __call__(self, args):
        objects = super(AddKeyCommand, self).__call__(args)
        NewKeyCommandBase.output(objects, 'add')


@commandlet("addcert")
class AddCert(Command):
    '''
    Adds a certificate object
    '''

    # adhere to an interface
    # pylint: disable=no-self-use
    def generate_options(self, group_parser):
        group_parser.add_argument(
            '--label', help='The profile label to remove.\n', required=True)

        group_parser.add_argument(
            'cert', help='The x509 PEM certificate to add.\n')

        sub_group = group_parser.add_mutually_exclusive_group()

        sub_group.add_argument(
            '--key-label',
            help='The associated private key label.\n')

        group_parser.add_argument(
            '--key-id',
            help='The associated private key id in hex.\n')


    def __call__(self, args):

        path = args['path']
        label = args['label']
        keylabel = args['key_label']
        keyid = args['key_id']
        certpath = args['cert']

        if (keylabel is None) == (keyid is None):
            sys.exit('Expected --key-label or --key-id to be specified')

        attrs = pemcert_to_attrs(certpath)

        pkcs11_object = PKCS11X509(attrs)

        with Db(path) as db:

            # get token to add to
            token = db.gettoken(label)

            # verify that key is existing
            # XXX we should be verifying that it's expected, but I guess one could always load up a cert
            # not associated with a key.
            tobjs = db.gettertiary(token['id'])

            # look up the private key
            missing_id_or_label = None
            for t in tobjs:
                if keylabel is not None:
                    missing_id_or_label = AddCert.get_id_by_label(t, keylabel)
                else:
                    missing_id_or_label = AddCert.get_label_by_id(t, keyid)
                if missing_id_or_label is not None:
                    break

            if missing_id_or_label is None:
                raise RuntimeError('Cannot find key with id "%s"' % keylabel)

            # have valid keylabel needed id
            if keylabel:
                pkcs11_object.update({CKA_ID: missing_id_or_label})
                pkcs11_object.update({CKA_LABEL: binascii.hexlify(keylabel.encode()).decode()})
            # have valid id needed keylabel
            else:
                pkcs11_object.update({CKA_LABEL: missing_id_or_label})
                pkcs11_object.update({CKA_ID: keyid})

            # TODO verify that cert is cryptographically bound to key found

            # add the cert
            db.addtertiary(token['id'], pkcs11_object)

        NewKeyCommandBase.output({'cert' : pkcs11_object}, 'add')

    @staticmethod
    def get_id_by_label(tobj, keylabel):

        attrs = yaml.safe_load(io.StringIO(tobj['attrs']))

        if CKA_LABEL in attrs:
            x = attrs[CKA_LABEL]
            x = binascii.unhexlify(x).decode()
            if x == keylabel and attrs[CKA_CLASS] == CKO_PRIVATE_KEY:
                return attrs[CKA_ID]

        return None

    @staticmethod
    def get_label_by_id(tobj, keyid):

        attrs = yaml.safe_load(io.StringIO(tobj['attrs']))

        if CKA_ID in attrs:
            x = attrs[CKA_ID]
            if x == keyid and attrs[CKA_CLASS] == CKO_PRIVATE_KEY:
                return attrs[CKA_LABEL] if CKA_LABEL in attrs else ''

        return None

@commandlet("objmod")
class ObjMod(Command):
    '''
    Dumps and modifies objects.
    '''

    _type_map = {
        'int' : 'do_int',
        'str' : 'do_str',
        'bool': 'do_bool',
        'raw' : 'do_raw',
    }

    @staticmethod
    def do_int(value):
        return int(value, 0)

    @staticmethod
    def do_bool(value):
        return str2bool(value)

    @staticmethod
    def do_str(value):
        return binascii.hexlify(value.encode()).decode()

    @staticmethod
    def do_raw(value):
        return value

    @classmethod
    def mod(cls, path, tid, key, value, inattrs, vtype):

        with Db(path) as db:
            obj = db.getobject(tid)
            if obj is None:
                sys.exit('Not found, object with id: {}'.format(tid))
        s = obj['attrs']
        obj_attrs = yaml.safe_load(s)

        # if we don't have any update data, just dump the attributes
        if not key and not inattrs:
            print(yaml.safe_dump(obj_attrs, default_flow_style=False))
            sys.exit()

        # if we have attributes YAML file, then we want to update all attributes
        if inattrs:
            with Db(path) as db:
                y = yaml.safe_load(open(inattrs, "r"))
                db.updatetertiary(obj['id'], y)
            sys.exit()

        # else we have --key and possibly --value
        #
        # look in the CKA_ globals from pkcs11t.py file for
        # a mapping string or raw value map.
        # filter(lambda x: x.startswith('CKA_'), globals().keys())
        keys = []
        for k in globals().keys():
            if k.startswith('CKA_'):
                keys.append(k)

        keynames = {}
        for k in keys:
            keynames[globals()[k]] = k

        keyname=None
        if key in keys:
            keyname=key
            key=globals()[key]
        else:
            key = int(key, 0)
            if key not in keynames:
                sys.exit('Unknown key: %d', key)
            keyname = keynames[key]

        if not value:
            if key and not key in obj_attrs:
                sys.exit("Key not found")

            print(yaml.safe_dump({keyname : obj_attrs[key]}))
            sys.exit()

        if not vtype:
            sys.exit("When specifying a value, type is required")

        value = getattr(cls, ObjMod._type_map[vtype])(value)
        obj_attrs[key] = value
        with Db(path) as db:
            db.updatetertiary(obj['id'], obj_attrs)

    # adhere to an interface
    def generate_options(self, group_parser):
        group_parser.add_argument(
            '--id', help='The object id.\n', required=True)
        group_parser.add_argument(
            '--key',
            help='The key to dump.\n')
        group_parser.add_argument(
            '--value',
            help='The value to set.\n')
        group_parser.add_argument(
            '--type',
            choices=self._type_map.keys(),
            help='Specify the type.\n')
        group_parser.add_argument(
            'attrs', nargs='?', help='The YAML attribute file.\n')
    def __call__(self, args):

        path = args['path']

        key = args['key'] if 'key' in args else None
        value = args['value'] if 'value' in args else None
        attrs = args['attrs'] if 'attrs' in args else None

        if attrs and key:
            sys.exit('Cannot specify --key when specifying the attributes')

        if attrs and value:
            sys.exit('Cannot specify --value when specifying the attributes')

        if value and not args['type']:
            sys.exit('require --type when specifying --value')

        ObjMod.mod(path, args['id'], key, value, attrs, args['type'])


@commandlet("objdel")
class ObjDel(Command):
    '''
    Deletes an object from a token.
    '''

    @classmethod
    def delete(cls, path, tid):

        with Db(path) as db:
            obj = db.getobject(tid)
            db.rmobject(obj['id'])

    # adhere to an interface
    def generate_options(self, group_parser):
        group_parser.add_argument(
            'id', help='The id of the object to delete.\n')

    def __call__(self, args):

        path = args['path']

        ObjDel.delete(path, args['id'])

@commandlet("link")
class LinkCommand(NewKeyCommandBase):
    '''
    Imports an existing TPM key to a token within a tpm2-pkcs11 store.
    '''

    # adhere to an interface
    # pylint: disable=no-self-use
    def generate_options(self, group_parser):
        super(LinkCommand, self).generate_options(group_parser)
        group_parser.add_argument('privkey',
            nargs='*',
            help='Path of the key to be linked.\n')
        group_parser.add_argument(
            '--auth',
            default='',
            help='The auth value for the key to link.\n'
        )

    def new_key_init(self, label, sopin, userpin, hierarchyauth, pobj, sealobjects, tpm2, d):

        pobj_handle = get_pobject(pobj, tpm2, hierarchyauth, d)

        # Get the primary object encrypted auth value and sokey information
        # to decode it. Based on the incoming pin
        is_so = sopin != None
        pin = sopin if is_so else userpin

        pubkey = '%spub' % ('so' if is_so else 'user')
        privkey = '%spriv' % ('so' if is_so else 'user')
        saltkey = '%sauthsalt' % ('so' if is_so else 'user')

        sealpub = sealobjects[pubkey]
        sealpriv = sealobjects[privkey]
        sealsalt = sealobjects[saltkey]

        sealctx = tpm2.load(pobj_handle, pobj['objauth'], sealpriv, sealpub)

        sealauth = hash_pass(pin, salt=sealsalt)['hash']

        wrappingkey = tpm2.unseal(sealctx, sealauth)

        wrapper = AESAuthUnwrapper(wrappingkey)

        objauth = self._auth

        encobjauth = wrapper.wrap(str2bytes(objauth))

        return (encobjauth, objauth)

    def create_from_tss_key(self, pobj, objauth, hierarchyauth, tpm2, alg, keypath, d):

        keypath = keypath[0]

        with open(keypath, "rb") as f:
            keybytes = f.read()
        tss2_privkey = TSSPrivKey.from_pem(keybytes)
        is_empty_auth = tss2_privkey.empty_auth
        phandle = tss2_privkey.parent
        pubbytes = tss2_privkey.public.marshal()
        privbytes = tss2_privkey.private.marshal()

        pid = pobj['id']
        pobj_config = yaml.safe_load(pobj['config'])
        is_transient = pobj_config['transient']
        if not is_transient and (phandle == tpm2.TPM2_RH_OWNER or \
            (phandle >> tpm2.TPM2_HR_SHIFT != tpm2.TPM2_HT_PERSISTENT)):
            sys.exit('The primary object (id: {:d}) is persistent and'
                ' the TSS Engine key does not have a persistent parent,'
                ' got: 0x{:x}'.format(pid, phandle))
        elif is_transient and not (phandle == tpm2.TPM2_RH_OWNER or phandle == 0):
            # tpm2-tss-engine < 1.1.0 used a phandle of 0 instead of tpm2.TPM2_RH_OWNER
            sys.exit('The primary object (id: {:d}) is transient and'
                ' the TSS Engine key has a parent handle,'
                ' got: 0x{:x}'.format(pid, phandle))


        if is_empty_auth and len(self._auth) if self._auth is not None else 0:
            sys.exit('Key expected to have auth value, please specify via option --auth');

        if not is_transient:
            # Im diving into the ESYS_TR serialized format,
            # this isn't the smartest thing to do...
            hexhandle = pobj_config['esys-tr']
            handle_bytes = binascii.unhexlify(hexhandle)[0:4]
            expected_handle = struct.unpack(">I", handle_bytes)[0]
            if phandle != expected_handle:
                sys.exit("Key must be parent of 0x{:X}, got 0x{:X}".format(
                    expected_handle, phandle))

        pobj_handle = get_pobject(pobj, tpm2, hierarchyauth, d)
        pobjauth = pobj['objauth']
        ctx = tpm2.load(pobj_handle, pobjauth, privbytes, pubbytes)
        tertiarypubdata, _ = tpm2.readpublic(ctx, False)

        privfd, tertiarypriv = mkstemp(prefix='', suffix='.priv', dir=d)
        try:
            os.write(privfd, privbytes)
        finally:
            os.close(privfd)

        pubfd, tertiarypub = mkstemp(prefix='', suffix='.pub', dir=d)
        try:
            os.write(pubfd, pubbytes)
        finally:
            os.close(pubfd)

        return (tertiarypriv, tertiarypub, tertiarypubdata)

    def create_from_key_blobs(self, pobj, objauth, hierarchyauth, tpm2, alg, keypaths, d):

        tertiarypub = keypaths[0]
        tertiarypriv = keypaths[1]

        pobj_handle = get_pobject(pobj, tpm2, hierarchyauth, d)
        pobjauth = pobj['objauth']
        try:
            ctx = tpm2.load(pobj_handle, pobjauth, tertiarypriv, tertiarypub)
        except RuntimeError as e:
            # Try swapping pub/priv
            tmp = tertiarypub
            tertiarypub = tertiarypriv
            tertiarypriv = tmp
            ctx = tpm2.load(pobj_handle, pobjauth, tertiarypriv, tertiarypub)

        tertiarypubdata, _ = tpm2.readpublic(ctx, False)

        return (tertiarypriv, tertiarypub, tertiarypubdata)

    # Links a new key
    def new_key_create(self, pobj, objauth, hierarchyauth, tpm2, alg, keypaths, passin, d):

        if keypaths is None:
            sys.exit("Keypath must be specified")

        if len(keypaths) == 1:
            return self.create_from_tss_key(pobj, objauth, hierarchyauth, tpm2, alg, keypaths, d)

        if len(keypaths) == 2:
            return self.create_from_key_blobs(pobj, objauth, hierarchyauth, tpm2, alg, keypaths, d)

        sys.exit("Expected one or two keyblobs, got: {}".format(len(keypaths)))

    def __call__(self, args):
        self._auth = args['auth'] if 'auth' in args else None
        objects = super(LinkCommand, self).__call__(args)
        NewKeyCommandBase.output(objects, 'link')


@commandlet("export")
class Export(Command):
    '''
    Exports an object from a token to a specified format.
    '''

    @staticmethod
    def _handle_tpm_key(db, obj, pin, is_so_pin, hierarchyauth, format, output_prefix):

        attrs = yaml.safe_load(io.StringIO(obj['attrs']))
        cka_class = attrs[CKA_CLASS]      

        if cka_class == CKO_SECRET_KEY:
            if format == "pem":
                raise RuntimeError(f'A format of "{format}" is not supported for CKO_SECRET_KEY objects')

            dump_blobs(db, obj, pin, is_so_pin, output_prefix)
            return

        if cka_class == CKO_PRIVATE_KEY:
            
            if format == "tpm2":           
                dump_blobs(db, obj, pin, is_so_pin, output_prefix)
            else:
                dump_tsspem(db, obj, pin, is_so_pin, output_prefix)               
            return
            
        raise RuntimeError(f'Object CKA_CLASS not supported, got: "{cka_class}"')
    
    @staticmethod
    def _handle_public_key(db, obj, pin, is_so_pin, hierarchyauth, format, output_prefix):
        dump_pubpem(db, obj, pin, is_so_pin, output_prefix)

    @staticmethod
    def export(db, tid, pin, is_so_pin, hierarchyauth, format, output_prefix):

        if not output_prefix:
            output_prefix = str(tid)

        obj = db.getobject(tid)
   
        attrs = yaml.safe_load(io.StringIO(obj['attrs']))
        
        cka_class = attrs[CKA_CLASS]

        handler = {
            CKO_SECRET_KEY : Export._handle_tpm_key,
            CKO_PRIVATE_KEY : Export._handle_tpm_key,
            CKO_PUBLIC_KEY : Export._handle_public_key,
        }
        
        if cka_class not in handler:
            raise RuntimeError(f'Object class not supported, got: "{cka_class}"')
        
       
        handler[cka_class](db, obj, pin, is_so_pin, hierarchyauth, format, output_prefix)
            
    def generate_options(self, group_parser):
        group_parser.add_argument(
            '--id', help='The id of the object to add (mutually exclusive with --label and --key-label).\n',
            type=int)

        group_parser.add_argument(
            '--label', help='The token label of the object to export. Requires --key-label\n')

        group_parser.add_argument(
            '--key-label', help='The label of the object to export. Required --label\n')

        group_parser.add_argument(
            '--format',
            choices = [ 'auto', 'tpm2', 'pem' ],
            default = 'auto',
            help='The format of the object to dump.\n')

        group_parser.add_argument(
            '--hierarchy-auth',
            help='The authorization password for the owner hierarchy when using a token with a transient primary object\n',
            default="")

        group_parser.add_argument(
            '--output-prefix',
            help='The output prefix. ' +
                'tpm2 blobs are output as suffix.priv and suffix.blob, pem outputs are suffix.pem\n')

        pinopts = group_parser.add_mutually_exclusive_group()
        pinopts.add_argument('--sopin', help='The Administrator pin.\n'),
        pinopts.add_argument('--userpin', help='The User pin.\n'),

    def __call__(self, args):

        object_id = args["id"]
        object_label = args["key_label"]
        token_label = args["label"]
        output_prefix = args['output_prefix']
        
        if object_id and object_label:
            sys.exit("Cannot specify --id with --key-label")

        elif object_id and token_label:
            sys.exit("Cannot specify --id with --label")

        elif not object_id and not token_label and not object_label:
            sys.exit("Must specify --label and --key-label or --id")

        path = args['path']

        is_so_pin = False

        pin = None        
        sopin = args['sopin']
        userpin = args['userpin']
        
        if sopin:
            pin = args['sopin']
            is_so_pin = True
        elif userpin:
            pin = args['userpin']

        with Db(path) as db:
            
            if not object_id:
                token = db.gettoken(label=token_label)
                objects = db.getobjects(token['id'])
                for obj in objects:
                    if AddCert.get_id_by_label(obj, object_label):
                        object_id = obj['id']
                        break
            
                if not object_id:
                    sys.exit(f'Could not find object with label "{object_label}" in token "{token_label}"')

                if not output_prefix:
                    output_prefix = object_label

            Export.export(db, object_id, pin, is_so_pin, args['hierarchy_auth'], args['format'], output_prefix)



# SPDX-License-Identifier: BSD-2-Clause
import argparse
import os
import textwrap

class commandlet(object):
    '''Decorator class for commandlet. You can add commandlets to the tool with this decorator.'''

    @staticmethod
    def get_default_store_path():

        # always use the env variable no matter what
        if "TPM2_PKCS11_STORE" in os.environ:
            store = os.environ.get("TPM2_PKCS11_STORE")
            try:
                os.mkdir(store, 0o770);
            except FileExistsError:
                return store
            except Exception:
                # Keep trying
                pass
            # Exists, use it
            return store

        # is their a system store and can I access it?
        store = "/etc/tpm2_pkcs11"
        if os.path.exists(store) and os.access(store, os.W_OK):
            return store

        # look for a store in home
        if "HOME" in os.environ:
            store = os.path.join(os.environ.get("HOME"), ".tpm2_pkcs11")
            try:
                os.mkdir(store, 0o770);
            except FileExistsError:
                return store
            except Exception:
                # Keep trying
                pass
            # Exists, use it
            return store

        # nothing else available, use cwd
        return os.getcwd()

    _commandlets = {}

    def __init__(self, cmd):
        self._cmd = cmd

        if cmd in commandlet._commandlets:
            raise Exception('Duplicate command name' + cmd)

        commandlet._commandlets[cmd] = None

    def __call__(self, cls):
        commandlet._commandlets[self._cmd] = cls()
        return cls

    @staticmethod
    def get():
        '''Retrieves the list of registered commandlets.'''
        return commandlet._commandlets

    @staticmethod
    def init(description):

        opt_parser = argparse.ArgumentParser(description=description)

        subparser = opt_parser.add_subparsers(help='commands')

        commandlets = commandlet.get()

        # for each commandlet, instantiate and set up their options
        for n, c in commandlets.items():
            p = subparser.add_parser(n, help=c.__doc__)
            p.set_defaults(which=n)
            # Instantiate

            opt_gen = getattr(c, 'generate_options', None)
            if callable(opt_gen):
                # get group help
                g = p.add_argument_group(n + ' options')
                # get args
                c.generate_options(g)
                g.add_argument(
                    '--path',
                    type=os.path.expanduser,
                    help=textwrap.dedent('''
                    The location of the store directory. If specified, the directory MUST exist.
                    If not specified performs a search by looking at environment variable
                    TPM2_PKCS11_STORE and if not set then /etc/tpm2_pkcs11 and if not found or
                    no write access, then $HOME/.tpm2_pkcs11 and if not found or cannot be created,
                    then defaults to using the current working directory.
                    '''),
                    default=commandlet.get_default_store_path())

        args = opt_parser.parse_args()

        d = vars(args)
        if 'which' in d:
            commandlet.get()[d['which']](d)
        else:
            opt_parser.print_usage()


class Command(object):
    '''Baseclass for a commandlet. Commandlets shall implement this interface.'''

    def generate_options(self, group_parser):
        '''Adds it's options to the group parser. The parser passed in is a result from
        calling add_argument_group(ArgumentGroup): https://docs.python.org/2/library/argparse.html
        Args:
            group_parser(): The parser to add options too.
        '''
        raise NotImplementedError('Implement: generate_options')

    def __call__(self, args):
        '''Called when the user selects your commandlet and passed the dictionary of arguments.
        Arguments:
            args ({str: arg}: The dictionary version of the attrs of the parser.
                The args value is obtained by:
                args = opt_parser.parse_args()
                args = vars(args)
                So to access args just do args['name']
        '''
        raise NotImplementedError('Implement: __call__')



# SPDX-License-Identifier: BSD-2-Clause
import binascii
import hashlib
import io
import os
import argparse
import sys
import shutil
import yaml
from tempfile import mkdtemp

from cryptography import exceptions as crypto_exceptions
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import (Cipher, algorithms, modes)
from cryptography.hazmat.primitives.asymmetric import (rsa, padding)
from cryptography.hazmat.primitives import hashes

from pyasn1_modules import pem, rfc2459
from pyasn1.codec.der import decoder as derdecoder, encoder as derencoder

from tpm2_pytss.ESAPI import ESAPI
from tpm2_pytss.tsskey import TSSPrivKey
from tpm2_pytss.constants import (
    TPM2_RH,
    ESYS_TR
)
from tpm2_pytss.types import (
    TPM2B_PUBLIC,
    TPM2B_PRIVATE,
)

from .pkcs11t import *  # noqa

def str2bytes(s):
    if isinstance(s, str):
        return s.encode()
    return s

def bytes_to_file(bites, tmpdir):
    path = os.path.join(tmpdir, "primary.handle")
    open(path, 'w+b').write(bites)
    return path

def rand_hex_str(num=32):
    if num % 2:
        raise RuntimeError("Expected even number of bytes, got: %u" % num)

    return binascii.hexlify(os.urandom(num // 2)).decode()


def hash_pass(password, salt=None):

    if salt is None:
        # get a 32 bit salt hex encoded (hex len 64)
        salt = rand_hex_str(64)

    # python 3.5.2 (doesn't seem to affect >= 3.5.6) is dumb...
    if isinstance(password, str):
        password = password.encode()

    if isinstance(salt, str):
        salt = salt.encode()

    m = hashlib.sha256(password)
    m.update(salt)

    # the TPM auth size is limited to 32 bytes in most cases
    digest = m.hexdigest()[:32]

    return {
        'salt': salt,
        'hash': digest,
    }


def query_yes_no(question, default="no"):
    """Ask a yes/no question via input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")


def load_sealobject(token, db, tpm2, pobjhandle, pobjauth, pin, is_so):

    sealobject = db.getsealobject(token['id'])
    if is_so:
        sealpub = sealobject['sopub']
        sealpriv = sealobject['sopriv']
        salt = sealobject['soauthsalt']
    else:
        sealpub = sealobject['userpub']
        sealpriv = sealobject['userpriv']
        salt = sealobject['userauthsalt']

    sealauth = hash_pass(pin.encode(), salt)['hash']

    # Load the so sealobject using the PARENTS AUTH (primaryobject)
    sealctx = tpm2.load(pobjhandle, pobjauth, sealpriv, sealpub)

    return sealctx, sealauth


class AESCipher:
    def __init__(self, key):
        self.key = key

    def encrypt(self, plaintext):
        # Generate a random 96-bit IV.
        iv = os.urandom(12)

        # Construct an AES-GCM Cipher object with the given key and a
        # randomly generated IV.
        encryptor = Cipher(
            algorithms.AES(self.key), modes.GCM(iv),
            backend=default_backend()).encryptor()

        ciphertext = encryptor.update(plaintext) + encryptor.finalize()

        iv = binascii.hexlify(iv).decode()
        ciphertext = binascii.hexlify(ciphertext).decode()
        tag = binascii.hexlify(encryptor.tag).decode()
        return ':'.join((iv, tag, ciphertext))

    def decrypt(self, ciphertext):

        iv, tag, ciphertext = ciphertext.split(':')
        iv = binascii.unhexlify(iv)
        tag = binascii.unhexlify(tag)
        ciphertext = binascii.unhexlify(ciphertext)

        decryptor = Cipher(
            algorithms.AES(self.key),
            modes.GCM(iv, tag),
            backend=default_backend()).decryptor()

        plaintext = decryptor.update(ciphertext) + decryptor.finalize()

        return plaintext


class TemporaryDirectory(object):
    """Context manager for tempfile.mkdtemp() so it's usable with "with" statement."""

    def __enter__(self):
        self.name = mkdtemp()
        return self.name

    def __exit__(self, exc_type, exc_value, traceback):
        shutil.rmtree(self.name)


class AESAuthUnwrapper(object):
    def __init__(self, key):

        self._cipher = AESCipher(binascii.unhexlify(key))

    def unwrap(self, value):
        return self._cipher.decrypt(value)

    def wrap(self, value):
        return self._cipher.encrypt(value)


# XXX
# TODO move to the argparse handler module
def str2bool(v):
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


def get_ec_params(alg):
    """Return a string representation of a hex encoded ASN1 object X9.62 EC parameter

    http://docs.oasis-open.org/pkcs11/pkcs11-curr/v2.40/cos01/pkcs11-curr-v2.40-cos01.html

    Indicates that EC parameters are byte arrays of a DER encoded ASN1 objects X9.62 parameter.
    This function will return a hex string without the leading 0x representing this encoding.
    """

    if alg == "NIST p256":
        obj = "2A8648CE3D030107"
    elif alg == "NIST p224":
        obj = "2B81040021"
    elif alg == "NIST p384":
        obj = "2B81040022"
    elif alg == "NIST p521":
        obj = "2B81040023"
    else:
        raise RuntimeError("alg %s has no EC params mapping" % alg)

    # start building the DER object tag + len + obj in hex
    der = "06{:02x}{}".format(len(obj) // 2, obj)

    return der


def asn1_format_ec_point_uncompressed(x, y):

    len_y = len(y)
    len_x = len(x)

    # normalize hex input to ensure no odd count hex strings
    if len_y % 2:
        len_y = len_y + 1
        y = "00" + y

    if len_x % 2:
        len_x = len_x + 1
        x = "00" + x

    # convert hex length to binary length
    len_y = len_y // 2
    len_x = len_x // 2

    # ensure that the binary representation fits into a byte
    total_len = len_y + len_x + 1
    if (total_len > 255):
        raise RuntimeError(
            "Length of X and Y plus uncompressed format byte greater than 255")

    # The uncompressed point format is:
    # <asn1 octet hdr> <len> <uncompressed point format byte> <X> <Y>
    # 04 <len> <X> <Y>

    s = "04{len:02x}04{X}{Y}".format(len=total_len, X=x, Y=y)

    return s

def pemcert_to_attrs(certpath):
    # rather than using pycryptography x509 parser, which gives native type access to certificate
    # fields use pyASN1 to get raw ASN1 encoded values for the fields as the spec requires them
    with open(certpath, "r") as f:
        bercert = pem.readPemFromFile(f)

    cert = derdecoder.decode(bercert, asn1Spec=rfc2459.Certificate())[0]
    c = cert['tbsCertificate']

    h = binascii.hexlify
    d = derencoder.encode

    hexbercert = h(bercert).decode()

    # the CKA_CHECKSUM attrs is the first 3 bytes of a sha1hash
    bercertchecksum = hashlib.sha1(bercert).digest()[0:3]
    hexbercertchecksum = h(bercertchecksum).decode()

    subj = c['subject']
    hexsubj = h(d(str2bytes(subj))).decode()

    issuer = c['issuer']
    hexissuer = h(d(str2bytes(issuer))).decode()

    serial = c['serialNumber']
    hexserial = h(d(str2bytes(serial))).decode()

    return {
        # The attrs of this attribute is derived by taking the first 3 bytes of the CKA_VALUE
        # field.
        CKA_CHECK_VALUE: hexbercertchecksum,
        # Start date for the certificate (default empty)
        CKA_START_DATE: "",
        # End date for the certificate (default empty)
        CKA_END_DATE: "",
        # DER-encoding of the SubjectPublicKeyInfo for the public key
        # contained in this certificate (default empty)
        CKA_PUBLIC_KEY_INFO: "",
        # DER encoded subject
        CKA_SUBJECT: hexsubj,
        # DER encoding of issuer
        CKA_ISSUER: hexissuer,
        # DER encoding of the cert serial
        CKA_SERIAL_NUMBER: hexserial,
        # BER encoding of the certificate
        CKA_VALUE: hexbercert,
        # RFC2279 string to URL where cert can be found, default empty
        CKA_URL: '',
        # hash of pub key subj, default empty
        CKA_HASH_OF_SUBJECT_PUBLIC_KEY: '',
        # Hash of pub key, default empty
        CKA_HASH_OF_ISSUER_PUBLIC_KEY: '',
        # Java security domain, default CK_SECURITY_DOMAIN_UNSPECIFIED
        CKA_JAVA_MIDP_SECURITY_DOMAIN: CK_SECURITY_DOMAIN_UNSPECIFIED,
        # Name hash algorithm, defaults to SHA1
        CKA_NAME_HASH_ALGORITHM: CKM_SHA_1
    }

def _pkcs11_to_str(value, prefix):

    g = globals()
    submap = dict(filter(lambda elem: elem[0].startswith(prefix), g.items()))
    inv_map = {v: k for k, v in submap.items()}

    if value in inv_map:
        return inv_map[value]
    else:
        return '0x{:X}'.format(value)

def pkcs11_cko_to_str(cko_value):

    return _pkcs11_to_str(cko_value, 'CKO_')

def pkcs11_ckk_to_str(ckk_value):

    return _pkcs11_to_str(ckk_value, 'CKK_')

def check_pss_signature(tpm2, pctx, pauth):

        (priv, pub, _) = tpm2.create(pctx, pauth=pauth)
        kctx = tpm2.load(pctx, pauth, priv, pub)
        (details, _) = tpm2.readpublic(kctx, False)
        y = yaml.safe_load(details)
        e = y['exponent'] if y['exponent'] != 0 else 65537
        n = int(y['rsa'], 16)

        pub_numbers = rsa.RSAPublicNumbers(e, n)
        public_key = pub_numbers.public_key(default_backend())

        message = b'message'
        signature = tpm2.sign(kctx, 'sha256', 'rsapss', message)

        try:
            public_key.verify(
                signature,
                message,
                padding.PSS(
                    mgf=padding.MGF1(hashes.SHA256()),
                    salt_length=hashes.SHA256().digest_size
                ),
                hashes.SHA256()
            )
            return True
        except crypto_exceptions.InvalidSignature:
            # saltLen = key_size_bytes - hash_size_bytes - 2;
            # If this fails, we're in some messed up position
            # PSS sigs are neither hlen == slen nor max(slen),
            # so we will fail loudly
            public_key.verify(
                signature,
                message,
                padding.PSS(
                    mgf=padding.MGF1(hashes.SHA256()),
                    salt_length=padding.PSS.MAX_LENGTH
                ),
                hashes.SHA256()
            )

            return False

def create_primary(tpm2, hierarchyauth, pobjauth, template=None):

        details = tpm2.TEMPLATES[template]
        alg = details['alg']
        attrs = details['attrs']
        
        if template == 'tss2-engine-key':
            supported = tpm2.getcap('algorithms')
            supported = yaml.safe_load(supported)
            if 'ecc' in supported:
                alg = 'ecc256'
            elif 'rsa' in supported:
                alg = 'rsa2048'
            else:
                sys.exit('TPM Supports neither RSA nor ECC')
        
        return tpm2.createprimary(hierarchyauth, pobjauth, alg=alg, attrs=attrs)

def get_pobject(pobject, tpm2, hierarchyauth, d):

    pobjconf = yaml.safe_load(pobject['config'])
    pobjauth = pobject['objauth']

    if pobjconf['transient'] is True:
        selection = pobjconf['template-name']
        pobj_handle = create_primary(tpm2, hierarchyauth, pobjauth, selection)
    else:
        tr_handle = binascii.unhexlify(pobjconf['esys-tr'])
        pobj_handle = bytes_to_file(tr_handle, d)

    return pobj_handle


def hierarchy_tpm_value(hierarchy):
    
    hierarchies = {
        'o' : TPM2_RH.OWNER
    }
           
    try:
        return hierarchies[hierarchy]
    except KeyError:
        raise RuntimeError(f'Hierarchy not supported, got: f{hierarchy}')

def getauth(db, obj, pin, is_sopin, hierarchyauth):
    
    from .tpm2 import Tpm2
    
    tokid = obj['tokid']
    pid = db.getpid_by_tokid(tokid)
    pobject = db.getprimary(pid)
    token = db.gettoken(id=tokid)

    attrs = yaml.safe_load(io.StringIO(obj['attrs']))
    
    with TemporaryDirectory() as d:
        tpm2 = Tpm2(d)
        
        token_config = yaml.safe_load(io.StringIO(token['config']))
        
        if pin is None:
            if token_config.get('empty-user-pin'):
                pin = ''
            else:
                sys.exit('error: at least one of the arguments --sopin --userpin is required')
        
        tobjauth=""
        pobj_handle = None
        encauth = None
        if CKA_TPM2_OBJAUTH_ENC in attrs:
            encauth = binascii.unhexlify(attrs[CKA_TPM2_OBJAUTH_ENC])

            pobjauth = pobject['objauth']

            pobj_handle = get_pobject(pobject, tpm2, hierarchyauth, d)

            sealctx, sealauth = load_sealobject(token, db, tpm2, pobj_handle, pobjauth,
                                                      pin, is_sopin)

            # get the ESYS_TR file into a state that can be used outside of the
            # temporary directory context tied to tpm2
            with open(pobj_handle, "rb") as f:
                pobj_handle = f.read()

            wrappingkey = tpm2.unseal(sealctx, sealauth)
        
            wrapper = AESAuthUnwrapper(wrappingkey)
                
            tobjauth=""
            if encauth:
                encauth=encauth.decode()
                tobjauth = wrapper.unwrap(encauth).decode()
        
        
        return (tobjauth, pobj_handle)

def _dump_outputs(objauth, pobj):
    
    config = yaml.safe_load(io.StringIO(pobj['config']))
    is_transient = 'esys-tr' not in config

    hierarchy = str(hierarchy_tpm_value(pobj['hierarchy']))

    output = {
        'object-auth' : f'{objauth}',
        'primary-object' : {
            'is_transient' : is_transient,
            'hierarchy' : hierarchy,
            'auth'      : pobj['objauth'] 
        }
    }
    
    document = yaml.safe_dump(output, default_flow_style=False)
    print(document)

def dump_blobs(db, obj, pin, is_sopin, output_prefix):

    tokid = obj['tokid']
    pid = db.getpid_by_tokid(tokid)
    pobj = db.getprimary(pid)

    attrs = yaml.safe_load(io.StringIO(obj['attrs']))

    pub_blob = TPM2B_PUBLIC.unmarshal(binascii.unhexlify(attrs[CKA_TPM2_PUB_BLOB]))[0]
    priv_blob = TPM2B_PRIVATE.unmarshal(binascii.unhexlify(attrs[CKA_TPM2_PRIV_BLOB]))[0]

    objauth = getauth(db, obj, pin, is_sopin, attrs)[0]    

    with open(output_prefix+ ".priv", "wb") as f:
        b = priv_blob.marshal()
        f.write(b)

    with open(output_prefix+ ".pub", "wb") as f:
        b = pub_blob.marshal()
        f.write(b)

    config = yaml.safe_load(io.StringIO(pobj['config']))
    is_transient = 'esys-tr' not in config
    if not is_transient:
        esys_tr = binascii.unhexlify(config['esys-tr'])
        with open(output_prefix+ ".tr", "wb") as f:
            f.write(esys_tr)

    _dump_outputs(objauth, pobj)

def dump_tsspem(db, obj, pin, is_sopin, output_prefix):

    tokid = obj['tokid']
    pid = db.getpid_by_tokid(tokid)
    pobj = db.getprimary(pid)

    attrs = yaml.safe_load(io.StringIO(obj['attrs']))

    pub_blob = TPM2B_PUBLIC.unmarshal(binascii.unhexlify(attrs[CKA_TPM2_PUB_BLOB]))[0]
    priv_blob = TPM2B_PRIVATE.unmarshal(binascii.unhexlify(attrs[CKA_TPM2_PRIV_BLOB]))[0]
    
    objauth, pobj_handle = getauth(db, obj, pin, is_sopin, attrs)    

    with ESAPI(os.getenv('TPM2TOOLS_TCTI', None)) as e:
        tr_handle = ESYS_TR.deserialize(e, pobj_handle)
        
        tpm_handle = e.tr_get_tpm_handle(tr_handle)

    key = TSSPrivKey(priv_blob, pub_blob, empty_auth=len(objauth) == 0,
                     parent=tpm_handle)

    with open(output_prefix + ".pem", "wb") as f:
        f.write(key.to_pem())

    _dump_outputs(objauth, pobj)

def dump_pubpem(db, obj, pin, is_sopin, output_prefix):
    
    attrs = yaml.safe_load(io.StringIO(obj['attrs']))
    
    pub_blob = TPM2B_PUBLIC.unmarshal(binascii.unhexlify(attrs[CKA_TPM2_PUB_BLOB]))[0]

    with open(output_prefix + ".pem", "wb") as f:
        f.write(pub_blob.to_pem())




# SPDX-License-Identifier: BSD-2-Clause
from __future__ import print_function
import os
import sys
from tempfile import mkstemp, NamedTemporaryFile
import uuid


from subprocess import Popen, PIPE

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePrivateKey
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey

from .utils import str2bytes

class Tpm2(object):
    TPM2_RH_OWNER = 0x40000001
    TPM2_HR_SHIFT = 24
    TPM2_HT_PERSISTENT = 0x81

    _TSS2_DEFAULT_ATTRS = "userwithauth|restricted|decrypt|noda|fixedtpm" \
        "|fixedparent|sensitivedataorigin"

    ALGS = [
        'rsa1024', 'rsa2048', 'rsa3072', 'rsa4096',
        'aes128', 'aes256',
        'ecc224', 'ecc256', 'ecc384', 'ecc521',
        'hmac:sha1', 'hmac:sha256', 'hmac:sha384', 'hmac:sha512'
    ]

    TEMPLATES = {
        # tpm2-tools default RSA2048 primary, same as tss2-engine-key
        None :  {
            'alg' : None,
            'attrs' : None
        },
        # tpm2-tools key compatible with RSA 2048, same as None
        'tpm2-tools-default' :  {
            'alg' : None,
            'attrs' : None
        },
        # tpm2-tools key compatible with EC P256
        'tpm2-tools-ecc-default' :  {
            'alg' : 'ecc',
            'attrs' : None
        },
        # tss2-engine key
        'tss2-engine-key' :  {
            'alg' : None,
            'attrs' : _TSS2_DEFAULT_ATTRS
        }
    }

    def __init__(self, tmp):
        self._tmp = tmp

    @property
    def tmpdir(self):
        return self._tmp

    def createprimary(self, hierarchyauth=None, objauth=None, alg=None, attrs=None,):
        ctx = os.path.join(self._tmp, "context.out")
        cmd = [
            'tpm2_createprimary', '-c', ctx, '-g',
            'sha256'
        ]

        if alg is None:
            alg = 'rsa'

        cmd.extend(['-G', alg])

        if objauth and len(objauth) > 0:
            cmd.extend(['-p', objauth])

        if hierarchyauth and len(hierarchyauth) > 0:
            cmd.extend(['-P', hierarchyauth])

        if attrs is not None:
            cmd.extend(['-a', attrs])

        p = Popen(cmd, stdout=PIPE, stderr=PIPE, env=os.environ)
        _, stderr = p.communicate()
        if (p.wait()):
            raise RuntimeError("Could not execute tpm2_createprimary: %s" %
                               stderr)
        return ctx

    def evictcontrol(self, hierarchyauth, ctx, handle=None):

        tr_file = os.path.join(self._tmp, "primary.handle")

        cmd = ['tpm2_evictcontrol', '-c', str(ctx), '-o', tr_file]

        if hierarchyauth and len(hierarchyauth) > 0:
            cmd.extend(['-P', hierarchyauth])

        if handle:
            cmd.append(str(handle))

        p = Popen(cmd, stdout=PIPE, stderr=PIPE, env=os.environ)
        stdout, stderr = p.communicate()
        if (p.wait()):
            raise RuntimeError("Could not execute tpm2_evictcontrol: %s" %
                               stderr)
        return tr_file

    def readpublic(self, handle, get_tr_file=True):

        tr_file = os.path.join(self._tmp, "primary.handle")

        cmd = ['tpm2_readpublic', '-c', str(handle)]

        if get_tr_file:
            cmd.extend(['-t', tr_file])

        p = Popen(cmd, stdout=PIPE, stderr=PIPE, env=os.environ)
        stdout, stderr = p.communicate()
        if (p.wait()):
            raise RuntimeError("Could not execute tpm2_readpublic: %s" %
                               stderr)
        return (stdout, tr_file if get_tr_file else None)

    def load(self, pctx, pauth, priv, pub):

        if priv != None and not isinstance(priv, str):
            sealprivf = NamedTemporaryFile()
            sealprivf.write(priv)
            sealprivf.flush()
            priv = sealprivf.name

        if not isinstance(pub, str):
            sealpubf = NamedTemporaryFile()
            sealpubf.write(pub)
            sealpubf.flush()
            pub = sealpubf.name

        ctx = os.path.join(self._tmp, uuid.uuid4().hex + '.out')
        #tpm2_load -C $file_primary_key_ctx  -u $file_load_key_pub  -r $file_load_key_priv -n $file_load_key_name -c $file_load_key_ctx
        if priv != None:
            cmd = [
                'tpm2_load', '-C', str(pctx), '-u', pub, '-r',
                priv, '-n', '/dev/null', '-c', ctx
            ]
            if pauth is not None:
                cmd.extend(['-P', pauth])
        else:
            cmd = ['tpm2_loadexternal', '-u', pub, '-c', ctx]

        p = Popen(cmd, stdout=PIPE, stderr=PIPE, env=os.environ)
        _, stderr = p.communicate()
        rc = p.wait()
        if rc:
            raise RuntimeError("Could not execute tpm2_load: %s" % stderr)
        return ctx

    def unseal(self, ctx, auth):

        # tpm2_unseal -Q -c $file_unseal_key_ctx
        cmd = ['tpm2_unseal', '-c', ctx, '-p', auth]
        p = Popen(cmd, stdout=PIPE, stderr=PIPE, env=os.environ)
        stdout, stderr = p.communicate()
        rc = p.wait()
        if rc:
            raise RuntimeError("Could not execute tpm2_unseal: %s" % stderr)
        return stdout

    def _encryptdecrypt(self, ctx, auth, data, decrypt=False):

        cmd = ['tpm2_encryptdecrypt', '-c', ctx, '-p', auth]

        if decrypt:
            cmd.extend(['-d'])

        p = Popen(cmd, stdout=PIPE, stderr=PIPE, stdin=PIPE, env=os.environ)
        stdout, stderr = p.communicate(input=data)
        rc = p.wait()
        if rc:
            raise RuntimeError("Could not execute tpm2_encryptdecrypt: %s" %
                               stderr)
        return stdout

    def encrypt(self, ctx, auth, data):
        return self._encryptdecrypt(ctx, auth, data)

    def decrypt(self, ctx, auth, data):
        return self._encryptdecrypt(ctx, auth, data, decrypt=True)

    def create(self,
               phandle,
               pauth=None,
               objauth=None,
               objattrs=None,
               seal=None,
               alg=None):
        # tpm2_create -Q -C context.out -g $gAlg -G $GAlg -u key.pub -r key.priv
        privfd, priv = mkstemp(prefix='', suffix='.priv', dir=self._tmp)
        os.close(privfd)
        pubfd, pub = mkstemp(prefix='', suffix='.pub', dir=self._tmp)
        os.close(pubfd)

        cmd = ['tpm2_create', '-C', str(phandle), '-u', pub, '-r', priv]

        if pauth and len(pauth) > 0:
            cmd.extend(['-P', '%s' % pauth])

        if objauth and len(objauth) > 0:
            cmd.extend(['-p', objauth])

        if objattrs != None:
            cmd.extend(['-a', objattrs])

        if seal != None:
            cmd.extend(['-i', '-'])

        if alg != None:
            cmd.extend(['-G', alg])

        p = Popen(cmd, stdout=PIPE, stderr=PIPE, stdin=PIPE, env=os.environ)
        stdout, stderr = p.communicate(input=str2bytes(seal))
        rc = p.wait()
        if (rc != 0):
            os.remove(pub)
            os.remove(priv)
            raise RuntimeError("Could not execute tpm2_create: %s" %
                               str(stderr))

        return priv, pub, stdout

    def getcap(self, cap):

        # tpm2_getcap -Q -l $cap
        cmd = ['tpm2_getcap', cap]
        p = Popen(cmd, stdout=PIPE, stderr=PIPE, env=os.environ)
        stdout, stderr = p.communicate()
        rc = p.wait()
        if rc:
            raise RuntimeError("Could not execute tpm2_getcap: %s" % stderr)
        return stdout

    def importkey(self,
                  phandle,
                  pauth,
                  objauth,
                  privkey,
                  objattrs=None,
                  seal=None,
                  alg=None,
                  passin=None):

        if privkey and len(privkey) > 0:
            exists = os.path.isfile(privkey)
            if not exists:
                raise RuntimeError("File '%s' path is invalid or is missing" %
                                   privkey)
        else:
            sys.exit("Invalid file path")

        privfd, priv = mkstemp(prefix='', suffix='.priv', dir=self._tmp)
        os.close(privfd)
        pubfd, pub = mkstemp(prefix='', suffix='.pub', dir=self._tmp)
        os.close(pubfd)

        # If the key is an OpenSSH key, convert it to PEM format
        pem_priv_name = None
        with open(privkey, "rb") as f:
            privey_data = f.read()
        if privey_data.startswith(b'-----BEGIN OPENSSH PRIVATE KEY-----'):
            if passin:
                # Parse passin to extract the password
                if passin.startswith('env:'):
                    password_bytes = os.getenvb(passin[4:].encode())
                elif passin.startswith('file:'):
                    with open(passin[5:], 'rb') as f:
                        password_bytes = f.read()
                elif passin.startswith('pass:'):
                    password_bytes = passin[5:].encode()
                else:
                    raise NotImplementedError("Unsupported OpenSSL password input {} to read OpenSSH key".format(
                        repr(passin)))
                enc_alg = serialization.BestAvailableEncryption(password_bytes)
            else:
                password_bytes = None
                enc_alg = serialization.NoEncryption()

            ssh_key = serialization.load_ssh_private_key(privey_data, password=password_bytes)

            if alg is None:
                # Find the algorithm
                if isinstance(ssh_key, EllipticCurvePrivateKey):
                    alg = 'ecc'
                elif isinstance(ssh_key, RSAPrivateKey):
                    alg = 'rsa'
                else:
                    raise NotImplementedError("Unsupported SSH key type {}".format(type(ssh_key)))

            pem_key = ssh_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=enc_alg)

            pem_priv_fd, pem_priv_name = mkstemp(prefix='', suffix='.privpem', dir=self._tmp)
            try:
                os.write(pem_priv_fd, pem_key)
            finally:
                os.close(pem_priv_fd)
            privkey = pem_priv_name
        elif alg is None:
            # Guess the key algorithm from the PEM header
            if privey_data.startswith(b'-----BEGIN EC PARAMETERS-----'):
                alg = 'ecc'
            elif privey_data.startswith(b'-----BEGIN EC PRIVATE KEY-----'):
                alg = 'ecc'
            elif privey_data.startswith(b'-----BEGIN RSA PRIVATE KEY-----'):
                alg = 'rsa'
            else:
                raise RuntimeError("Unable to detect key type, use --algorithm to specify it")

        parent_path = str(phandle)
        cmd = [
            'tpm2_import', '-V', '-C', parent_path, '-i', privkey, '-u', pub,
            '-r', priv
        ]

        if pauth and len(pauth) > 0:
            cmd.extend(['-P', pauth])

        if objauth and len(objauth) > 0:
            cmd.extend(['-p', objauth])

        if objattrs != None:
            cmd.extend(['-a', objattrs])

        if seal != None:
            cmd.extend(['-i', '-'])

        if alg != None:
            cmd.extend(['-G', alg])

        if passin is not None:
            cmd.extend(['--passin', passin])

        p = Popen(cmd, stdout=PIPE, stderr=PIPE, stdin=PIPE, env=os.environ)
        stdout, stderr = p.communicate(input=seal)
        rc = p.wait()
        if pem_priv_name is not None:
            os.remove(pem_priv_name)
        if (rc != 0):
            os.remove(pub)
            os.remove(priv)
            print("command: %s" % str(" ".join(cmd)))
            raise RuntimeError("Could not execute tpm2_import: %s" %
                               stderr)

        return priv, pub, stdout

    def changeauth(self, pctx, objctx, oldobjauth, newobjauth):

        newpriv = os.path.join(self._tmp, uuid.uuid4().hex + '.priv')

        cmd = [
            'tpm2_changeauth',
            '-C',
            str(pctx),
            '-c',
            str(objctx),
            '-p',
            oldobjauth,
            '-r',
            newpriv,
            newobjauth,
        ]
        p = Popen(cmd, stdout=PIPE, stderr=PIPE, env=os.environ)
        _, stderr = p.communicate()
        rc = p.wait()
        if rc:
            raise RuntimeError("Could not execute tpm2_changeauth: %s" % stderr)

        return newpriv

    def sign(self, ctx, halg, scheme, message):

        sig = os.path.join(self._tmp, uuid.uuid4().hex + '.dat')

        cmd = [
            'tpm2_sign',
            '-c',
            str(ctx),
            '-g', halg,
            '-s', scheme,
            '-f',
            'plain',
            '-o',
            sig
        ]

        p = Popen(cmd, stdout=PIPE, stderr=PIPE, stdin=PIPE, env=os.environ)
        stdout, stderr = p.communicate(input=str2bytes(message))
        rc = p.wait()
        if (rc != 0):
            print("command: %s" % str(" ".join(cmd)))
            raise RuntimeError("Could not execute tpm2_import: %s" %
                               str(stderr))
        data = open(sig, "rb").read()
        os.unlink(sig)
        return data
